/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: BitImage.c,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $";

#include <math.h>
#include <stdlib.h>

#include <vim/proto.h>

#include <vim/BitImageP.h>
#include <vim/CMManage.h>

extern double rint __((double));

static void
  Initialize __((GenericImage im, ArgList argv, Cardinal argc));
static void
  Destroy __((GenericImage im));

static void
  Redisplay __((GenericDisplayParams disp, Boolean same));

BitImageClassRec bitImageClassRec = {
  { /* core fields */
    /* superclass		*/	&genericImageClassRec,
    /* class_name		*/	"BitImage",
    /* image_size		*/	sizeof(BitImageRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* resources                */	NULL,
    /* num_resources            */	0,
    /* set_values               */      NULL,
    /* read			*/	NULL,
    /* write			*/	NULL,
    /* image_display_size	*/	sizeof(BitDisplayParamsRec),
    /* param_resources          */	NULL,
    /* num_param_resources      */	0,
    /* initialize_params	*/	NULL,
    /* set_params		*/	NULL,
    /* destroy_params		*/	NULL,
    /* edit_params		*/	NULL,
    /* remember_params		*/	NULL,
    /* compare_params		*/	NULL,
    /* install			*/	NULL,
    /* uninstall		*/	NULL,
    /* alloc_colors             */      NULL,
    /* free_colors              */      NULL,
    /* display			*/	Redisplay,
    /* description		*/	NULL,
    /* element_description	*/	NULL,
    /* destroy			*/      Destroy
  }
};

ImageClass bitImageClass = (ImageClass) &bitImageClassRec;

static void
  Initialize _IMPL((gim, argv, argc),
		   GenericImage gim _AND ArgList argv _AND Cardinal argc )
{
    BitImage im = (BitImage) gim;
    int width  = im->generic.width;
    int height = im->generic.height;
    int linelen;

    linelen = (width / 8) + (width % 8 ? 1 : 0);
    im->bit.data = (byte *) XtCalloc(height, linelen);
}

static void
  Destroy _IMPL((gim), GenericImage gim )
{
    BitImage im = (BitImage) gim;

    XtFree((char*) im->bit.data);
}

static byte *
zoomBitImage _DCL((xmag, ymag, bits, bit_width, bit_height, 
		   xp, yp, widthp, heightp), 
		  float xmag _AND float ymag _AND 
		  byte *bits _AND int bit_width _AND int bit_height _AND
		  int *xp _AND int *yp _AND int *widthp _AND int *heightp );

static void
  Redisplay _IMPL((im_disp, same),
		  GenericDisplayParams im_disp _AND Boolean same )
{
    BitImage bit = (BitImage) im_disp->generic.image;
    ExternalParams *xparams = &(im_disp->xparams);
    Display *disp = xparams->display;
    Window win = xparams->win;
    Visual *vis = xparams->visual;
    float mag_x = xparams->mag_x;
    float mag_y = xparams->mag_y;
    Region region = xparams->region;
    XRectangle *range = &xparams->range;
    int ddepth = DefaultDepthOfScreen(xparams->screen);
    int src_x, src_y, dst_x, dst_y, width, height;
    XImage *xim = 0;
    byte *zim = 0;
    GC gc; XGCValues gcv;
    use_BitImage(bit,bit);
    Pixmap pm = im_disp->generic.pm;

    if ( !xparams ) abort();

    dst_x = (int) rint(mag_x*range->x);
    dst_y = (int) rint(mag_y*range->y);
    width = im_disp->generic.pm_width;
    height= im_disp->generic.pm_height;

    /* Create an image on the given range. */
    if ( !same || !pm ) {
	if ( pm ) XFreePixmap(disp,pm);
	width  = range->width;
	height = range->height;
	if ( mag_x == 1.0 && mag_y == 1.0 ) {
	    xim = XCreateImage(disp, vis, 1, XYBitmap,
			       0, (char*) bit_data, bit_width, bit_height,
			       8, (bit_width+7)/8 );
	    src_x = range->x;
	    src_y = range->y;
	} else {
	    src_x = range->x;
	    src_y = range->y;
	    zim = zoomBitImage(mag_x, mag_y, 
			       bit_data, bit_width, bit_height,
			       &src_x, &src_y, &width, &height );
	    xim = XCreateImage(disp, vis, 1, XYBitmap,
			       0, (char*) zim, width, height, 
			       8, (width+7)/8 );
	}
	xim->bitmap_bit_order = MSBFirst;
	xim->byte_order = MSBFirst;
    }

    /* Create the appropriate GC. */
    gcv.function = GXcopy;
    gcv.foreground = CMblackPixel;
    gcv.background = CMwhitePixel;
    gc = XCreateGC( disp, win, GCFunction|GCBackground|GCForeground, &gcv );

    if ( xim ) {
	/* Finally draw it. */
	im_disp->generic.pm = pm = 
	  XCreatePixmap(disp, win, width, height, ddepth );
	im_disp->generic.pm_width  = width;
	im_disp->generic.pm_height = height;
	XPutImage( disp, pm, gc, xim, src_x, src_y, 0, 0, width, height );
    }

    XSetRegion(disp, gc, region);
    XCopyArea(disp, pm, win, gc, 0, 0, width, height, dst_x, dst_y);
    
    /* And clean up. */
    if ( zim ) 	free(zim);
    if ( xim ) 	free(xim);
    XFreeGC(disp,gc);
}

static unsigned int *
  buildIndex _IMPL((x, width, mag, rx, rwidth),
		   int x _AND int width _AND float mag _AND 
		   int *rx _AND int *rwidth)
{
    unsigned int *index;
    unsigned int  a;

    *rwidth = (int) rint(mag * width);
    *rx = 0;

    index = (unsigned int *) malloc(*rwidth * sizeof(unsigned int));

    for (a = *rx; a < *rwidth; a++)
      index[a] = a/mag + x ;
    return(index);
}

static byte *
zoomBitImage _IMPL((xmag, ymag, bits, bit_width, bit_height, 
		   xp, yp, widthp, heightp),
		   float xmag _AND float ymag _AND 
		   byte *bits _AND int bit_width _AND int bit_height _AND
		   int *xp _AND int *yp _AND int *widthp _AND int *heightp )
{
  unsigned int *xindex, *yindex;
  unsigned int  x, y, xsrc, ysrc, pysrc, tysrc;
  unsigned int  srclinelen;
  unsigned int  destlinelen;
  byte         *dest;
  byte         *srcline, *psrcline, *srcptr, *tsrcptr;
  byte         *destline, *destptr;
  byte          srcmask, destmask, bit;
  unsigned int  srcwidth  = *widthp;
  unsigned int  srcheight = *heightp;
  unsigned int  srcx = *xp;
  unsigned int  srcy = *yp;

  xindex = buildIndex(srcx, srcwidth,  xmag, xp, widthp);
  yindex = buildIndex(srcy, srcheight, ymag, yp, heightp);

  destlinelen = (*widthp + 7)/8;
  dest = destline = (byte *) calloc( *heightp * destlinelen, 1 );

  srclinelen = (bit_width + 7)/8;
  srcline = bits + srcy*srclinelen;
  
  for (y = 0, ysrc = yindex[y]; y < *heightp; y++) {
      pysrc = ysrc;
      psrcline = srcline;
      while (ysrc != yindex[y]) {
	  ysrc++;
	  srcline += srclinelen;
      }
      srcptr   = srcline + (srcx/8);
      srcmask  = 0x80 >> (srcx%8);
      destptr  = destline;
      destmask = 0x80;
      bit = srcmask & *srcptr;

      for (x = 0, xsrc = xindex[x]; x < *widthp; x++) {
	  int tbits = 0, bbits = 0;
	  if (xsrc != xindex[x]) {
	      do {
		  xsrc++;
		  if (!(srcmask >>= 1)) {
		      srcmask= 0x80;
		      srcptr++;
		  }
		  bit = srcmask & *srcptr;
		  for ( tysrc = ysrc, tsrcptr = srcptr; 
		        tysrc > pysrc;
		        tysrc--, tsrcptr -= srclinelen ) {
		      bit = srcmask & *tsrcptr;
		      tbits++; if (bit) ++bbits;
		  };
	      } while (xsrc != xindex[x]);
	      if ( tbits>1 ) bit = (2*bbits>tbits);
	  }
	  if (bit) *destptr |= destmask;
	  if (!(destmask >>= 1)) {
	      destmask = 0x80;
	      destptr++;
	  }
      }
      destline += destlinelen;
  }

  free((char *)xindex);
  free((char *)yindex);

  return dest;
}
