/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

   $Id: ChooseVar.h,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $

******************************************************************************/

#ifndef _ChooseVar_h
#define _ChooseVar_h

#include <vim/proto.h>

/****************************************************************
 *
 * ChooseVar widget
 *
 ****************************************************************/

/* Resources:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 background	     Background		Pixel		XtDefaultBackground
 border		     BorderColor	Pixel		XtDefaultForeground
 borderWidth	     BorderWidth	Dimension	1
 choices             Choices            Choices         NULL
 destroyCallback     Callback		Pointer		NULL
 font		     Font		XFontStruct	XtDefaultFont
 foreground	     Foreground		Pixel		XtDefaultForeground
 height		     Height		Dimension	0
 mappedWhenManaged   MappedWhenManaged	Boolean		True
 selectFont	     BoldFont		XFontStruct	XtDefaultFont
 sensitive	     Sensitive		Boolean		True
 unselectFont	     Font		XFontStruct	XtDefaultFont
 width		     Width		Dimension	0
 x		     Position		Position	0
 y		     Position		Position	0
*/

/* define any special resource names here that are not in <X11/StringDefs.h> */
#define XtNcvTitle "cvTitle"
#define XtNchoices "choices"
#define XtNselectFont "selectFont"
#define XtNunselectFont "unselectFont"

#define XtCChoices "Choices"
#define XtCBoldFont "BoldFont"

#define XtRChoices "Choices"

/* declare specific ChooseVarWidget class and instance datatypes */

typedef struct _ChooseVarWidgetClassRec*	ChooseVarWidgetClass;
typedef struct _ChooseVarWidgetRec*		ChooseVarWidget;

/* declare the class constant */

extern WidgetClass chooseVarWidgetClass;

/* Now declare functions for manipulating these. */

typedef void (*change_f) __((Widget w, void *old, void *new));

typedef struct {
    String name, prompt;
    XrmQuark type;
    int idx;
    void *var;
    change_f changep;
    XtPointer *args;
    void *cv_data;
    Widget cv_widget;
    Widget widget;
} cvChoice;

typedef struct {
    String name, type;
    int idx;
    String prompt;
    void *var;
    change_f changep;
    XtPointer *args;
} cvChoiceDescr;

#define cvSetChoice( arg, _n, _t, _i, _p, _v, _c, _a )		\
  ((arg).name = (_n), (arg).type = (_t), (arg).idx = (_i),	\
   (arg).prompt = (_p), (arg).var = (_v), (arg).changep = (_c),	\
   (arg).args = (_a), 0)

#define cvGeneric(arg, var, type, idx, prompt, changep, args) 	\
  cvSetChoice(arg, _STRING(var), _STRING(type), (idx), (prompt),\
	      (void*) &(var), (change_f)(changep), (args))

#define cvSpace(arg, pxls) 					\
  cvSetChoice(arg, "space", "Space", (pxls), 0, 0, 0, 0)	\

#define cvButton(arg, name, callback, prompt, cdata ) 		\
  cvSetChoice(arg, (name), "Button", 0, (prompt),		\
	      (void*)(cdata), (change_f)(callback), 0)

#define cvbEndButtons "!EndButtons"
#define cvbSameLine   "!SameLine"

#define cvBoolean(arg, var, idx, prompt, changep, args) 	\
  cvSetChoice(arg, _STRING(var), "Boolean", (idx), (prompt),	\
	      (void*) &(var), (change_f)(changep), (args))

#define cvChar(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "Char", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvUChar(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "UChar", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvShort(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "Short", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvUShort(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "UShort", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvInt(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "Int", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvUInt(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "UInt", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvFloat(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "Float", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvDouble(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "Double", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvString(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "String", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvEnum(arg, var, idx, prompt, changep, args) 		\
  cvSetChoice(arg, _STRING(var), "Enum", (idx), (prompt),	\
	      (void*)&(var), (change_f)(changep), (args))

#define cvEnumPair(str, val) (XtPointer)(str), (XtPointer)(val)

extern XtTranslations cvReadTrans;
extern XtTranslations cvEditTrans;

Boolean
  cvRegisterType __((String type, int idx, 
		     Widget (*cf)(Widget, cvChoice *),
		     void (*initf)(cvChoice *, Boolean),
		     void (*resetf)(cvChoice *),
		     void (*changef)(Widget),
		     Boolean (*changedf)(cvChoice *)
		     ));
void
  cvSetChoices __((Widget cv, cvChoiceDescr *descr, Cardinal num));

Boolean
  ChooseVarModal __((ChooseVarWidget shell, XtGrabKind kind, Boolean placed));

/* These functions are useful for implementing your own choice types. */
ChooseVarWidget
  findCVWidget __((Widget w));
cvChoice *
  getCVChoice __((Widget w));

/* Callbacks for standard exit buttons.  Callable from user callbacks. */
void cvDone __((Widget w, XtPointer closure, XtPointer call_data));
void cvRestore __((Widget w, XtPointer closure, XtPointer call_data));
void cvCancel __((Widget w, XtPointer closure, XtPointer call_data));

Widget cvCreateAsciiTextWidget __((String name, Widget parent));

#endif /* _Image_h */
