/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

   $Id: GenImage.h,v 0.9.1.3 1991/12/16 20:56:36 leei Exp $

******************************************************************************/

#ifndef _GenImage_h_
#define _GenImage_h_

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <vim/proto.h>

#if defined(__GNUC__) && !defined(alloca)
#define alloca __builtin_alloca
#elif defined(sun)
#include <alloca.h>
#else
extern char *alloca __((int));
#endif

#define ImNwidth XtNwidth
#define ImNheight XtNheight
#define ImNident "ident"
#define ImNvisible "visible"

#define ImCWidth XtCWidth
#define ImCHeight XtCHeight
#define ImCIdent "Ident"
#define ImCVisible "Visible"

/* declare specific GenericImage class and instance datatypes */

typedef struct _GenericImageClassRec*		ImageClass;
typedef struct _GenericImageRec*		GenericImage;
typedef struct _GenericDisplayParamsRec*	GenericDisplayParams;

typedef struct _ExternalParams {
    Widget widget;
    Display *display;
    Screen *screen;
    Visual *visual;
    Drawable win;
    Region region;
    XRectangle range;
    float mag_x, mag_y;
} ExternalParams;

typedef struct {
    String name;
    String type;
    XrmValue value;
} ImProp, *ImPropList;

#define ImSetProp( arg, n, t, d ) \
  ((void) ((arg).name = (n), (arg).type = (t), \
	   (arg).value.size = sizeof(d), (arg).value.addr = (caddr_t)(d)))

/* declare the class constant */

extern ImageClass genericImageClass;

/* declare the externally visible function call interface. */

ImageClass
  ImClass __((GenericImage im));
Boolean
  ImIsSubclass __((GenericImage im, ImageClass cl));
_CONST char *
  ImIdent __((GenericImage im));
int
  ImWidth __((GenericImage im));
int
  ImHeight __((GenericImage im));
Boolean
  ImChanged __((GenericImage im));

_CONST char *
  ImComment __((GenericImage im));
_CONST char *
  ImSetComment __((GenericImage im, _CONST char *comment));
_CONST char *
  ImAppendComment __((GenericImage im, _CONST char *comment, Boolean timestamp));

void
  ImPutProperties __((GenericImage im, ImPropList propv, Cardinal propc));
Boolean
  ImGetProperties __((GenericImage im, ImPropList propv, Cardinal propc));
Boolean
  ImConvert __((XrmQuark type1, XrmValuePtr val1, 
		XrmQuark type2, XrmValuePtr val2));

ImageClass
  ImClassSuperclass __((ImageClass class));
_CONST char *
  ImClassName __((ImageClass class));

GenericImage
  ImCreateImage __((char *id, ImageClass class, int w, int h,
		    ArgList argv, Cardinal argc));
void
  ImUpdateImage __((GenericImage im, XRectangle *range));

GenericDisplayParams
  ImCreateParams __((GenericImage im, ExternalParams *xparams));
void
  ImDestroyParams __((GenericDisplayParams disp, Boolean destroy_image));
ExternalParams *
  ImExternalParams __((GenericDisplayParams disp));

void
  ImRemoveImage __((GenericImage image));
void
  ImDestroyImage __((GenericImage im));
char *
  ImDescription __((GenericImage im, char *buf, int buf_len));
char *
  ImElementDescription __((GenericImage im, int i, int j, 
			   char *buf, int buf_len));
void
  ImAllocColors __((GenericDisplayParams disp));
void
  ImFreeColors __((GenericDisplayParams disp));
void
  ImDisplay __((GenericDisplayParams disp));
Boolean
  ImSetParams __((GenericDisplayParams disp, ArgList argv, Cardinal argc));
Boolean
  ImEditParams __((GenericDisplayParams disp));
void
  ImRememberParams __((GenericDisplayParams disp));
Boolean
  ImSameParams __((GenericDisplayParams disp));

void
  ImInstall __((GenericDisplayParams disp));
void
  ImUninstall __((GenericDisplayParams disp));

/* Declare imFileList access and manipulation. */
typedef struct {
    char *filename;
    GenericImage image;
} ImFileAssocRec, *ImFileAssoc;

extern ImFileAssoc imFileList;
extern int imFileListLength;

void ImAddFile __((GenericImage image));
ImFileAssoc ImFindFile __((char *filename));

/* Declare ImActionList registration and access. */
typedef void (*ImActionFunc) __((GenericDisplayParams, Widget, XEvent *, 
				 String *, Cardinal *));
typedef struct {
    String name;
    ImActionFunc func;
} ImAction, *ImActionList;

void
  ImAddActions __((ImActionList actions, Cardinal num_actions));

#endif /* _GenImage_h_ */
