/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

   $Id: GenImageP.h,v 0.9.1.4 1991/12/19 13:38:21 leei Exp $

******************************************************************************/

#ifndef _GenImageP_h_
#define _GenImageP_h_

/* 
 * This section defines the base of a class heirarchy similar to the 
 * Core widget heirarchy, except that it is a class of 2D images.
 */

#include <stdio.h>

#include <vim/GenImage.h>
#include <vim/proto.h>

#undef visible			/* Problem with MPL */

typedef struct {
    struct _GenericImageClassRec *superclass;
    char *class_name;
    unsigned int image_size;
    void (*class_initialize)
      __((ImageClass class));
    void (*class_part_initialize)
      __((ImageClass class));
    int class_inited;
    void (*initialize)
      __((GenericImage im, ArgList argv, Cardinal argc));
    XtResourceList resources;
    Cardinal num_resources;
    Boolean (*set_values)
      __((GenericImage current, GenericImage request, GenericImage new));
    void (*read)
      __((GenericImage im, FILE *file));
    void (*write)
      __((GenericImage im, FILE *file));
    unsigned int image_display_size;
    XtResourceList param_resources;
    Cardinal num_param_resources;
    void (*initialize_params)
      __((GenericDisplayParams disp));
    Boolean (*set_params)
      __((GenericDisplayParams disp, 
	  GenericDisplayParams request, 
	  GenericDisplayParams new));
    void (*destroy_params)
      __((GenericDisplayParams disp));
    void (*edit_params)
      __((GenericDisplayParams disp));
    void (*remember_params)
      __((GenericDisplayParams disp));
    Boolean (*same_params)
      __((GenericDisplayParams disp));
    void (*install)
      __((GenericDisplayParams disp));
    void (*uninstall)
      __((GenericDisplayParams disp));
    int (*alloc_colors)
      __((GenericDisplayParams disp));
    void (*free_colors)
      __((GenericDisplayParams disp));
    void (*display)
      __((GenericDisplayParams disp, Boolean same));
    void (*description)
      __((GenericImage im, char *buf, int buf_len));
    void (*element_description)
      __((GenericImage im, int i, int j, char *buf, int buf_len));
    void (*destroy)
      __((GenericImage im));
} GenericImageClassPart;

typedef struct _GenericImageClassRec {
    GenericImageClassPart generic;
} GenericImageClassRec;

extern GenericImageClassRec genericImageClassRec;

typedef struct _ImageProperty {
    struct _ImageProperty *next;
    XrmQuark name;
    XrmQuark type;
    XrmValue data;
} *ImageProperty;

typedef struct {
    /* Required fields. */
    ImageClass class;
    unsigned short width;
    unsigned short height;
    char *ident;
    /* Optional fields. */
    char *comment;
    ImageProperty props;
    char *file_name;
    char *file_type;
    int (*file_reader) __((GenericImage im, FILE *fp));
    int (*file_writer) __((GenericImage im, FILE *fp));
    time_t timestamp;
    int changed;
    GenericDisplayParams params;
} GenericImagePart;

typedef struct _GenericImageRec {
    GenericImagePart generic;
} GenericImageRec;

typedef struct {
    GenericImage image;
    GenericDisplayParams prev;
    GenericDisplayParams next;
    Pixmap pm;
    Dimension pm_width;
    Dimension pm_height;
    Boolean visible;
} GenericDisplayParamsPart;

typedef struct _GenericDisplayParamsRec {
    ExternalParams xparams;
    GenericDisplayParamsPart generic;
} GenericDisplayParamsRec;

#endif /* _GenImageP_h_ */
