/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: ImViewport.c,v 0.9.1.2 1991/11/17 19:28:02 leei Exp $";

/***********************************************************
Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#include <stdio.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/ShellP.h>

#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xmu/Misc.h>

#include <X11/Xmu/Converters.h>

#include "Scrollbar.h"
#include "EzMenu.h"

#include "Image.h"
#include "ImViewportP.h"

static void ScrollUpDownProc(), ThumbProc();
static Boolean GetGeometry();

#define offset(field) XtOffset(ImageViewportWidget,viewport.field)
static XtResource resources[] = {
    {XtNallowHoriz, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(allowhoriz), XtRImmediate, (caddr_t)True},
    {XtNallowVert, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(allowvert), XtRImmediate, (caddr_t)True},
    {XtNuseBottom, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(usebottom), XtRImmediate, (caddr_t)True},
    {XtNuseRight, XtCBoolean, XtRBoolean, sizeof(Boolean),
	 offset(useright), XtRImmediate, (caddr_t)True},
    {XtNmovePercent, XtCPercent, XtRInt, sizeof(int),
	 offset(movePercent), XtRImmediate, (caddr_t) 80 },
    {XtNimage, XtCImage, XtRWidget, sizeof(Widget),
	 offset(image), XtRImmediate, (caddr_t)NULL}
    
};
#undef offset

static char translations[] =
"     <Motion>:        imShowMenuBar() \n\
";

static void Initialize(), ConstraintInitialize(),
    Realize(), Resize(), ChangeManaged();
static Boolean SetValues(), Layout();
static XtGeometryResult GeometryManager(), PreferredGeometry();

#define superclass	(&formClassRec)
ImageViewportClassRec imageViewportClassRec = {
  { /* core_class fields */
    /* superclass	  */	(WidgetClass) superclass,
    /* class_name	  */	"ImageViewport",
    /* widget_size	  */	sizeof(ImageViewportRec),
    /* class_initialize	  */	XawInitializeWidgetSet,
    /* class_part_init    */    NULL,
    /* class_inited	  */	FALSE,
    /* initialize	  */	Initialize,
    /* initialize_hook    */    NULL,
    /* realize		  */	Realize,
    /* actions		  */	NULL,
    /* num_actions	  */	0,
    /* resources	  */	resources,
    /* num_resources	  */	XtNumber(resources),
    /* xrm_class	  */	NULLQUARK,
    /* compress_motion	  */	TRUE,
    /* compress_exposure  */	TRUE,
    /* compress_enterleave*/    TRUE,
    /* visible_interest	  */	FALSE,
    /* destroy		  */	NULL,
    /* resize		  */	Resize,
    /* expose		  */	XtInheritExpose,
    /* set_values	  */	SetValues,
    /* set_values_hook    */    NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */	NULL,
    /* accept_focus	  */	NULL,
    /* version            */    XtVersion,
    /* callback_private	  */	NULL,
    /* tm_table    	  */	translations,
    /* query_geometry     */    PreferredGeometry,
    /* display_accelerator*/	XtInheritDisplayAccelerator,
    /* extension          */	NULL
  },
  { /* composite_class fields */
    /* geometry_manager	  */	GeometryManager,
    /* change_managed	  */	ChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension          */	NULL
  },
  { /* constraint_class fields */
    /* subresourses	  */	NULL,
    /* subresource_count  */	0,
    /* constraint_size	  */	sizeof(ImageViewportConstraintsRec),
    /* initialize	  */	ConstraintInitialize,
    /* destroy		  */	NULL,
    /* set_values	  */	NULL,
    /* extension          */	NULL
  },
  { /* form_class fields */
    /* layout		  */	Layout
  },
  { /* viewport_class fields */
    /* empty		  */	0
  }
};

WidgetClass imageViewportWidgetClass = (WidgetClass)&imageViewportClassRec;

static Widget CreateScrollbar(w, horizontal)
    ImageViewportWidget w;
    Boolean horizontal;
{
    Widget clip = w->viewport.clip;
    ImageViewportConstraints constraints =
	(ImageViewportConstraints)clip->core.constraints;
    Arg barArgs[8]; Cardinal nargs;
    Widget bar;

    nargs = 0;
    XtSetArg(barArgs[nargs], XtNorientation,
	     horizontal ? XtorientHorizontal: XtorientVertical );
    ++nargs;
    XtSetArg(barArgs[nargs], XtNlength,
	     horizontal ? clip->core.width: clip->core.height);
    ++nargs;
    XtSetArg(barArgs[nargs], XtNleft,
	     (!horizontal&&w->viewport.useright) ? XtChainRight: XtChainLeft);
    ++nargs;
    XtSetArg(barArgs[nargs], XtNright,
	     (!horizontal&&!w->viewport.useright) ? XtChainLeft: XtChainRight);
    ++nargs;
    XtSetArg(barArgs[nargs], XtNtop,
	     (horizontal&&w->viewport.usebottom) ? XtChainBottom: XtChainTop);
    ++nargs;
    XtSetArg(barArgs[nargs], XtNbottom,
	     (horizontal&&!w->viewport.usebottom) ? XtChainTop: XtChainBottom);
    ++nargs;
    XtSetArg(barArgs[nargs], XtNcentre, True );
    ++nargs;
    XtSetArg(barArgs[nargs], XtNmappedWhenManaged, False );
    ++nargs;

    bar = XtCreateWidget( (horizontal ? "horizontal" : "vertical"),
			  ukcScrollbarWidgetClass, (Widget)w,
			  barArgs, nargs );
    XtAddCallback( bar, XtNscrollProc, ScrollUpDownProc, (caddr_t)w );
    XtAddCallback( bar, XtNjumpProc, ThumbProc, (caddr_t)w );

    if (horizontal) {
	w->viewport.horiz_bar = bar;
	constraints->form.vert_base = bar;
    }
    else {
	w->viewport.vert_bar = bar;
	constraints->form.horiz_base = bar;
    }

    XtManageChild( bar );

    return bar;
}

/* ARGSUSED */
static void Initialize(request, new)
    Widget request, new;
{
    ImageViewportWidget w = (ImageViewportWidget)new;
    static Arg args[5];
    Cardinal num_args;
    Widget h_bar, v_bar, menu_bar, box, button, menu;
    Dimension clip_height, clip_width;

    w->form.default_spacing = 0;  /* Reset the default spacing to 0 pixels. */

/* 
 * Initialize scrollbar widget pointers to NULL.
 */
    w->viewport.horiz_bar = w->viewport.vert_bar = (Widget)NULL;

/* 
 * Create menu bar widget.
 */
    w->viewport.menu_bar = menu_bar =
      XtVaCreateWidget("menuBar", boxWidgetClass, new,
		       XtNmappedWhenManaged, False,
		       NULL);

    button = 
      XtVaCreateManagedWidget("imageButton", menuButtonWidgetClass, menu_bar,
			      XtNmenuName, "imageMenu",
			      NULL);
    menu = 
      XtVaCreatePopupShell("imageMenu", ezMenuWidgetClass, button, 
			   NULL);

    button = 
      XtVaCreateManagedWidget("fileButton", menuButtonWidgetClass, menu_bar,
			      XtNmenuName, "fileMenu",
			      NULL);
    menu = 
      XtVaCreatePopupShell("fileMenu", ezMenuWidgetClass, button, 
			   NULL);

    button = 
      XtVaCreateManagedWidget("stackButton", menuButtonWidgetClass, menu_bar,
			      XtNmenuName, "stackMenu",
			      NULL);
    menu = 
      XtVaCreatePopupShell("stackMenu", ezMenuWidgetClass, button, 
			   NULL);

    button = 
      XtVaCreateManagedWidget("userButton", menuButtonWidgetClass, menu_bar,
			      XtNmenuName, "userMenu",
			      NULL);
    menu = 
      XtVaCreatePopupShell("userMenu", ezMenuWidgetClass, button, 
			   NULL);
    
/*
 * Initialize size of viewport based on size of image.
 */
    if ( w->viewport.image ) {
	int image_w = w->viewport.image->core.width;
	int image_h = w->viewport.image->core.height;
	int image_bw = w->viewport.image->core.border_width;
	
	if ( !w->core.width  ) w->core.width  = image_w + 2*image_bw;
	if ( !w->core.height ) w->core.height = image_h + 2*image_bw;
    }

/* 
 * Create Clip Widget.
 */
    num_args = 0;
    XtSetArg(args[num_args], XtNborderWidth, 0); num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft); num_args++;
    XtSetArg(args[num_args], XtNright, XtChainRight); num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop); num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainBottom); num_args++;

    w->viewport.clip = XtCreateManagedWidget("clip", widgetClass, new,
					     args, num_args);


    if (w->viewport.allowhoriz) 
        (void) CreateScrollbar(w, True);
    if (w->viewport.allowvert) 
        (void) CreateScrollbar(w, False);

    h_bar = w->viewport.horiz_bar;
    v_bar = w->viewport.vert_bar;

/*
 * Set the clip widget to the correct height.
 */

    clip_width = w->core.width;
    clip_height = w->core.height;

    if ( (h_bar != NULL) &&
	 (w->core.width > h_bar->core.width + h_bar->core.border_width) )
        clip_width -= h_bar->core.width + h_bar->core.border_width;
    
    if ( (v_bar != NULL) &&
	 (w->core.height > v_bar->core.height + v_bar->core.border_width) )
        clip_height -= v_bar->core.height + v_bar->core.border_width;

    num_args = 0;
    XtSetArg(args[num_args], XtNwidth, clip_width); num_args++;
    XtSetArg(args[num_args], XtNheight, clip_height); num_args++;
    XtSetValues(w->viewport.clip, args, num_args);

    /* Initialize the action translations (prior to realization) */
    AddImageActions(new);
}

/* ARGSUSED */
static void ConstraintInitialize(request, new)
    Widget request, new;
{
    ((ImageViewportConstraints)new->core.constraints)
      ->viewport.reparented = False;
}

static void Realize(widget, value_mask, attributes)
    Widget widget;
    XtValueMask *value_mask;
    XSetWindowAttributes *attributes;
{
    ImageViewportWidget w = (ImageViewportWidget)widget;
    register Widget image = w->viewport.image;
    register Widget clip = w->viewport.clip;
    register Widget menu_bar = w->viewport.menu_bar;

    *value_mask |= CWBitGravity;
    attributes->bit_gravity = NorthWestGravity;
    (*superclass->core_class.realize)(widget, value_mask, attributes);

    (*w->core.widget_class->core_class.resize)(widget);	/* turn on bars */

    if (image != (Widget)NULL) {
	XtMoveWidget( image, (Position)0, (Position)0 );
	XtRealizeWidget( clip );
	XtRealizeWidget( image );
	XReparentWindow( XtDisplay(w), XtWindow(image), XtWindow(clip),
			 (Position)0, (Position)0 );
	XtMapWidget( image );
	XtSetKeyboardFocus(clip, image);
    }

    XtInstallAllAccelerators(widget, widget);

    /* Initialize the menu_bar position. */
    XtRealizeWidget(menu_bar);
}

/* ARGSUSED */
static Boolean SetValues(current, request, new)
    Widget current, request, new;
{
    ImageViewportWidget w = (ImageViewportWidget)new;

    /* take care of bars, &tc. */
    (*w->core.widget_class->core_class.resize)( new );
    return False;
}


static void ChangeManaged(widget)
    Widget widget;
{
    ImageViewportWidget w = (ImageViewportWidget)widget;
    register int num_children = w->composite.num_children;
    register Widget image, *imageP;
    register int i;

    image = (Widget)NULL;
    for (imageP=w->composite.children, i=0; i < num_children; imageP++, i++) {
	if (XtIsManaged(*imageP)
	    && *imageP != w->viewport.clip
	    && *imageP != w->viewport.horiz_bar
	    && *imageP != w->viewport.vert_bar)
	{
	    image = *imageP;
	    break;
	}
    }

    if (image != w->viewport.image) {
	w->viewport.image = image;
	if (image != (Widget)NULL) {
	    XtInstallAllAccelerators(image, widget);
	    XtResizeWidget( image, image->core.width,
			    image->core.height, (Dimension)0 );
	    if (XtIsRealized(widget)) {
		ImageViewportConstraints constraints =
		    (ImageViewportConstraints)image->core.constraints;
		if (!XtIsRealized(image)) {
		    Window window = XtWindow(w);
		    XtMoveWidget( image, (Position)0, (Position)0 );
		    w->core.window = XtWindow(w->viewport.clip);
		    XtRealizeWidget( image );
		    w->core.window = window;
		    constraints->viewport.reparented = True;
		}
		else if (!constraints->viewport.reparented) {
		    XReparentWindow( XtDisplay(w), XtWindow(image),
				     XtWindow(w->viewport.clip),
				     (Position)0, (Position)0 );
		    constraints->viewport.reparented = True;
		    if (image->core.mapped_when_managed)
			XtMapWidget( image );
		}
	    }
	    GetGeometry( widget, 
			image->core.width  + 2*image->core.border_width, 
			image->core.height + 2*image->core.border_width );
	    (*((ImageViewportWidgetClass)w->core.widget_class)->form_class.layout)
		( (FormWidget)w, w->core.width, w->core.height );
	    /* %%% do we need to hide this image from Form?  */
	}
    } else if ( !image ) {
	GetGeometry(widget, 
		    w->core.width<=0?  200: w->core.width,
		    w->core.height<=0? 200: w->core.height);
	XtResizeWidget(w->viewport.clip,w->core.width,w->core.height,0);
    }

#ifdef notdef
    (*superclass->composite_class.change_managed)( widget );
#endif
}


static void SetBar(w, top, length, total)
    Widget w;
    Position top;
    Dimension length, total;
{
    XukcScrollbarSetThumb(w, (float) top / total, (float) length / total );
}

static void RedrawThumbs(w)
  ImageViewportWidget w;
{
    register Widget image = w->viewport.image;
    register Widget clip = w->viewport.clip;

    if (w->viewport.horiz_bar != (Widget)NULL)
	SetBar( w->viewport.horiz_bar, -(image->core.x),
	        clip->core.width, image->core.width );

    if (w->viewport.vert_bar != (Widget)NULL)
	SetBar( w->viewport.vert_bar, -(image->core.y),
	        clip->core.height, image->core.height );
}



static void MoveImage(w, x, y)
    ImageViewportWidget w;
    Position x, y;
{
    register Widget image = w->viewport.image;
    register Widget clip  = w->viewport.clip;

    int image_width  = image->core.width  + 2*image->core.border_width;
    int image_height = image->core.height + 2*image->core.border_width;

    if ( image_width >= clip->core.width ) {
	/* make sure we never move past right/bottom borders */
	if (-x + clip->core.width > image_width)
	  x = -(image_width - clip->core.width);
	if ( x > 0 ) x = 0;
    } else {
	x = (clip->core.width - image_width)/2;
    }

    if ( image_height >= clip->core.height ) {
	if (-y + clip->core.height > image_height)
	  y = -(image_height - clip->core.height);
	if ( y > 0 ) y = 0;
    } else {
	y = (clip->core.height - image_height)/2;
    }

    XtMoveWidget(image, x, y);

    RedrawThumbs(w);
}

static void ComputeLayout(widget, query, destroy_scrollbars)
    Widget widget;		/* ImageViewport */
    Boolean query;		/* query image's preferred geom? */
    Boolean destroy_scrollbars;	/* destroy un-needed scrollbars? */
{
    ImageViewportWidget w = (ImageViewportWidget)widget;
    register Widget image = w->viewport.image;
    register Widget clip = w->viewport.clip;
    ImageViewportConstraints constraints
	= (ImageViewportConstraints)clip->core.constraints;
    Boolean needshoriz, needsvert;
    int clip_width, clip_height;
    int image_width, image_height;
    XtWidgetGeometry intended;
    Dimension prev_width, prev_height;
    XtGeometryMask prev_mode;
    XtWidgetGeometry preferred;

    if (image == (Widget) NULL) return;

    clip_width = w->core.width;
    clip_height = w->core.height;

    image_width  = image->core.width  + 2*image->core.border_width;
    image_height = image->core.height + 2*image->core.border_width;

    /* Sanity check to prevent CreateScrollbars blowup. */
    if ( clip->core.width == 0 || clip->core.height == 0 )
      XtResizeWidget( clip, (Dimension)clip_width,
		     (Dimension)clip_height, (Dimension)0 );

    intended.request_mode = 0;
    needshoriz = needsvert = False;

    /*
     * intended.{width,height} caches the eventual image dimensions,
     * but we don't set the mode bits until after we decide that the
     * image's preferences are not acceptable.
     */

    if (!w->viewport.allowhoriz) 
      intended.request_mode |= CWWidth;

	if (image_width < clip_width) 
	    intended.width = clip_width;
	else
	    intended.width = image->core.width;

	if (image_height < clip_height) 
	    intended.height = clip_height;
	else
	    intended.height = image->core.height;

	if (!w->viewport.allowvert) 
	    intended.request_mode |= CWHeight;

	if (!query) {
	    preferred.width  = image->core.width;
	    preferred.height = image->core.height;
	}
	do { /* while intended != prev  */

	    if (query) {
	        (void) XtQueryGeometry( image, &intended, &preferred );
		if ( !(preferred.request_mode & CWWidth) )
		    preferred.width = intended.width;
		if ( !(preferred.request_mode & CWHeight) )
		    preferred.height = intended.height;
	    }
	    prev_width = intended.width;
	    prev_height = intended.height;
	    prev_mode = intended.request_mode;

	    /*
	     * note that having once decided to turn on either bar
	     * we'll not change our mind until we're next resized,
	     * thus avoiding potential oscillations.
	     */
#define CheckHoriz()							  \
	    if (w->viewport.allowhoriz && preferred.width > clip_width) { \
	        if (!needshoriz) {					  \
		    Widget bar;					          \
		    needshoriz = True;				          \
		    if ((bar = w->viewport.horiz_bar) == (Widget)NULL)    \
		        bar = CreateScrollbar(w, True);		          \
		    clip_height -= bar->core.height +		          \
			           bar->core.border_width;		  \
		    if (clip_height < 1) clip_height = 1;		  \
		}							  \
	        intended.width = preferred.width;			  \
	    }
/*enddef*/
	    CheckHoriz();
	    if (w->viewport.allowvert && preferred.height > clip_height) {
	        if (!needsvert) {
		    Widget bar;
		    needsvert = True;
		    if ((bar = w->viewport.vert_bar) == (Widget)NULL)
		        bar = CreateScrollbar(w, False);
		    clip_width -= bar->core.width + bar->core.border_width;
		    if (clip_width < 1) clip_width = 1;
		    CheckHoriz();
		}
		intended.height = preferred.height;
	    }
	    if (!w->viewport.allowhoriz || preferred.width < clip_width) {
	        intended.width = clip_width;
		intended.request_mode |= CWWidth;
	    }
	    if (!w->viewport.allowvert || preferred.height < clip_height) {
	        intended.height = clip_height;
		intended.request_mode |= CWHeight;
	    }
	} while (intended.request_mode != prev_mode
		 || (intended.request_mode & CWWidth
		     && intended.width != prev_width)
		 || (intended.request_mode & CWHeight
		     && intended.height != prev_height));

    if (XtIsRealized(clip))
	XRaiseWindow( XtDisplay(clip), XtWindow(clip) );

    XtMoveWidget( clip,
		  needsvert ? (w->viewport.useright ? 0 :
			       w->viewport.vert_bar->core.width +
			       w->viewport.vert_bar->core.border_width) : 0,
		  needshoriz ? (w->viewport.usebottom ? 0 :
				w->viewport.horiz_bar->core.height +
			        w->viewport.horiz_bar->core.border_width) : 0);
    XtResizeWidget( clip, (Dimension)clip_width,
		    (Dimension)clip_height, (Dimension)0 );

    if (w->viewport.horiz_bar != (Widget)NULL) {
	register Widget bar = w->viewport.horiz_bar;
	if (!needshoriz) {
	    constraints->form.vert_base = (Widget)NULL;
	    if (destroy_scrollbars) {
		XtDestroyWidget( bar );
		w->viewport.horiz_bar = (Widget)NULL;
	    }
	}
	else {
	    register int bw = bar->core.border_width;
	    XtResizeWidget( bar, clip_width, bar->core.height, bw );
	    XtMoveWidget( bar,
			  ((needsvert && !w->viewport.useright)
			   ? w->viewport.vert_bar->core.width
			   : -bw),
			  (w->viewport.usebottom
			    ? w->core.height - bar->core.height - bw
			    : -bw) );
	    XtSetMappedWhenManaged( bar, True );
	}
    }

    if (w->viewport.vert_bar != (Widget)NULL) {
	register Widget bar = w->viewport.vert_bar;
	if (!needsvert) {
	    constraints->form.horiz_base = (Widget)NULL;
	    if (destroy_scrollbars) {
		XtDestroyWidget( bar );
		w->viewport.vert_bar = (Widget)NULL;
	    }
	}
	else {
	    register int bw = bar->core.border_width;
	    XtResizeWidget( bar, bar->core.width, clip_height, bw );
	    XtMoveWidget( bar,
			  (w->viewport.useright
			   ? w->core.width - bar->core.width - bw 
			   : -bw),
			  ((needshoriz && !w->viewport.usebottom)
			    ? w->viewport.horiz_bar->core.height
			    : -bw) );
	    XtSetMappedWhenManaged( bar, True );
	}
    }

    if (image != (Widget)NULL) {
	MoveImage(w,
		     (needshoriz ? 
		      image->core.x :
		      (clip->core.width - image_width)/2),
		     (needsvert ?
		      image->core.y :
		      (clip->core.height - image_height)/2) );
    }
}

static void Resize(widget)
    Widget widget;
{
    ComputeLayout( widget, /*query=*/True, /*destroy=*/True );
}


/* ARGSUSED */
static Boolean Layout(w, width, height)
    FormWidget w;
    Dimension width, height;
{
    ComputeLayout( (Widget)w, /*query=*/True, /*destroy=*/True );
    w->form.preferred_width  = w->core.width;
    w->form.preferred_height = w->core.height;
    return False;
}


static void ScrollUpDownProc(widget, closure, call_data)
    Widget widget;
    caddr_t closure;
    caddr_t call_data;
{
    ImageViewportWidget w = (ImageViewportWidget)closure;
    register Widget image = w->viewport.image;
    int pix = (int)call_data;
    Position x, y;

    if (image == NULL) return;	/* no image to scroll. */

    x = image->core.x - ((widget == w->viewport.horiz_bar) ? pix : 0);
    y = image->core.y - ((widget == w->viewport.vert_bar) ? pix : 0);
    MoveImage(w, x, y);
}


/* ARGSUSED */
static void ThumbProc(widget, closure, percent)
    Widget widget;
    caddr_t closure;
    float *percent;
{
    ImageViewportWidget w = (ImageViewportWidget)closure;
    register Widget image = w->viewport.image;
    Position x, y;

    if (image == NULL) return;	/* no image to scroll. */

    if (widget == w->viewport.horiz_bar)
#ifdef macII				/* bug in the macII A/UX 1.0 cc */
	x = (int)(-*percent * image->core.width);
#else /* else not macII */
	x = -(int)(*percent * image->core.width);
#endif /* macII */
    else
	x = image->core.x;

    if (widget == w->viewport.vert_bar)
#ifdef macII				/* bug in the macII A/UX 1.0 cc */
	y = (int)(-*percent * image->core.height);
#else /* else not macII */
	y = -(int)(*percent * image->core.height);
#endif /* macII */
    else
	y = image->core.y;

    MoveImage(w, x, y);
}

static XtGeometryResult GeometryManager(image, request, reply)
    Widget image;
    XtWidgetGeometry *request, *reply;
{
    ImageViewportWidget w = (ImageViewportWidget)image->core.parent;
    int rMode = request->request_mode;
    Boolean rWidth  = (Boolean)(rMode & CWWidth);
    Boolean rHeight = (Boolean)(rMode & CWHeight);
    Boolean rBorderWidth = (Boolean)(rMode & CWBorderWidth);
    Boolean rX = (Boolean)(rMode & CWX);
    Boolean rY = (Boolean)(rMode & CWY);

    if (image != w->viewport.image || !(rWidth|rHeight|rX|rY|rBorderWidth) )
      return XtGeometryNo;

    GetGeometry( (Widget)w,
		(rWidth? 
		 request->width + 2*image->core.border_width: 
		 w->core.width),
		(rHeight? 
		 request->height + 2*image->core.border_width:
		 w->core.height)
		);

    if ( (rWidth && image->core.width != request->width) 
	|| (rHeight && image->core.height != request->height) 
	|| (rX && image->core.x != request->x)
	|| (rY && image->core.y != request->y)
	|| (rBorderWidth&& image->core.border_width!=request->border_width) ) {
	/* Compute the center of the window in its own coordinates. */
	
	if ( rX ) image->core.x = request->x;
	if ( rY ) image->core.y = request->y;
	if ( rBorderWidth ) image->core.border_width = request->border_width;

	{
	    int center_x = w->core.width/2
	      - image->core.x - image->core.border_width;
	    int center_y = w->core.height/2
	      - image->core.y - image->core.border_width;
	
	    if ( rWidth || rBorderWidth ) {
		if ( !rX )
		  image->core.x = w->core.width/2 - image->core.border_width -
		    (center_x*request->width)/image->core.width;
		image->core.width  = request->width;
	    }
	    if ( rHeight || rBorderWidth ) {
		if ( !rY )
		  image->core.y = w->core.height/2 - image->core.border_width -
		    (center_y*request->height)/image->core.height;
		image->core.height = request->height;
	    }
	}

	if ( rWidth|rHeight|rBorderWidth ) {
	    ComputeLayout( (Widget)w, /*query=*/ False, /*destroy=*/ True );
	} else {
	    MoveImage( (Widget)w, image->core.x, image->core.y );
	}
    }

    return XtGeometryYes;
}


static Boolean GetGeometry(w, width, height)
    Widget w;
    Dimension width, height;
{
    XtWidgetGeometry geometry;
    XtGeometryResult result;

    if (width == w->core.width && height == w->core.height)
	return False;

    geometry.request_mode = CWWidth | CWHeight;
    geometry.width = width;
    geometry.height = height;

    if (XtIsRealized(w)) {
	if ( (((ImageViewportWidget)w)->viewport.allowhoriz && 
	      width > w->core.width) 
	    || width < w->core.width )
	  geometry.width = w->core.width;
	if ( (((ImageViewportWidget)w)->viewport.allowvert && 
	      height > w->core.height)
	    || height < w->core.height )
	  geometry.height = w->core.height;
    } else {
	/* This is the Realize call; we'll inherit a w&h iff none currently */
	if (w->core.width != 0) {
	    if (w->core.height != 0) return False;
	    geometry.width = w->core.width;
	}
	if (w->core.height != 0) geometry.height = w->core.height;
    }

    result = XtMakeGeometryRequest(w, &geometry, &geometry);
    if ( result == XtGeometryAlmost )
      result = XtMakeGeometryRequest(w, &geometry, NULL);

    return (result == XtGeometryYes);
}

static XtGeometryResult PreferredGeometry(w, constraints, reply)
    Widget w;
    XtWidgetGeometry *constraints, *reply;
{
    if (((ImageViewportWidget)w)->viewport.image != NULL)
	return XtQueryGeometry( ((ImageViewportWidget)w)->viewport.image,
			       constraints, reply );
    else
	return XtGeometryYes;
}

void
actResizeToImage _IMPL((ww, ev, str, card),
		       Widget ww _AND XEvent *ev _AND 
		       String *str _AND Cardinal *card )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    Widget image;
    Dimension width, height;

    if ( !w ) return;
    
    image  = w->viewport.image;
    width  = image->core.width  + 2*image->core.border_width;
    height = image->core.height + 2*image->core.border_width;

    if ( XtMakeResizeRequest( (Widget)w, width, height, &width, &height )
	 == XtGeometryAlmost )
      XtMakeResizeRequest( (Widget)w, width, height, NULL, NULL );
    ComputeLayout((Widget) w, /*query*/False, /*destroy*/True);
}

void
  actPageLeft _IMPL((ww, ev, str, card),
		    Widget ww _AND XEvent *ev _AND
		    String *str _AND Cardinal *card )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    Widget clip, image;
    Dimension dist;

    if ( !w ) return;

    clip = w->viewport.clip;
    image = w->viewport.image;
    dist = (w->viewport.movePercent*clip->core.width)/100;
    
    if ( image ) MoveImage( w, image->core.x + dist, image->core.y );
}

void
actPageRight _IMPL((ww, ev, str, card),
		   Widget ww _AND XEvent *ev _AND
		   String *str _AND Cardinal *card )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    Widget clip, image;
    Dimension dist;

    if ( !w ) return;

    clip = w->viewport.clip;
    image = w->viewport.image;
    dist = (w->viewport.movePercent*clip->core.width)/100;
    
    if ( image ) MoveImage( w, image->core.x - dist, image->core.y );
}

void
actPageUp _IMPL((ww, ev, str, card),
		Widget ww _AND XEvent *ev _AND
		String *str _AND Cardinal *card )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    Widget clip, image;
    Dimension dist;

    if ( !w ) return;

    clip = w->viewport.clip;
    image = w->viewport.image;
    dist = (w->viewport.movePercent*clip->core.height)/100;
    
    if ( image ) MoveImage( w, image->core.x, image->core.y + dist );
}

void
actPageDown _IMPL((ww, ev, str, card),
		  Widget ww _AND XEvent *ev _AND
		  String *str _AND Cardinal *card )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    Widget clip, image;
    Dimension dist;

    if ( !w ) return;

    clip = w->viewport.clip;
    image = w->viewport.image;
    dist = (w->viewport.movePercent*clip->core.height)/100;
    
    if ( image ) MoveImage( w, image->core.x, image->core.y - dist );
}
