/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: Image.c,v 0.9.1.5 1991/12/20 14:22:00 leei Exp $";

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>

#include "ChooseVar.h"

#include "GenImageP.h"
#include "ImageP.h"
#include "ImViewportP.h"

#include "readImage.h"
#include "CMManage.h"
#include "proto.h"
#include "utils.h"
#include "debug.h"

#include <stdlib.h>
#include <math.h>

static int grid_spacings[] = {
    1,  1,  2,  5, 10,  10,  10,  20,  50,  50, 100,  100,  200
  };
static int grid_highlights[] = {
    5, 10, 10, 10, 100, 50, 100, 100, 500, 100, 1000, 500, 1000
  };

static _CONST float one = 1.0;

static XtResource resources[] = {
#define offset(field) XtOffset(ImageWidget, image.field)
    /* {name, class, type, size, offset, default_type, default_addr}, */
    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(foreground_pixel), XtRString, "XtDefaultForeground" },
    { XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(descr_font), XtRString, "XtDefaultFont" },
    { XtNgridFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(grid_font), XtRString, "XtDefaultFont" },
    { XtNmagX, XtCMag, XtRFloat, sizeof(float),
	offset(mag_x), XtRFloat, (XtPointer) &one },
    { XtNmagY, XtCMag, XtRFloat, sizeof(float),
	offset(mag_y), XtRFloat, (XtPointer) &one },
    { XtNdescription, XtCDescription, XtRBoolean, sizeof(Boolean),
	offset(descr), XtRImmediate, (XtPointer) True },
    { XtNgrid, XtCGrid, XtRBoolean, sizeof(Boolean),
	offset(grid), XtRImmediate, (XtPointer) False },
    { XtNverbose, XtCVerbose, XtRBoolean, sizeof(Boolean),
	offset(verbose), XtRImmediate, (XtPointer) False },
    { XtNredrawsDelayed, XtCRedrawsDelayed, XtRBoolean, sizeof(Boolean),
	offset(redraws_delayed), XtRImmediate, (XtPointer) False },
    { XtNdisplayN, XtCDisplayN, XtRInt, sizeof(int),
	offset(displayN), XtRImmediate, (XtPointer) 100 },
    { XtNstackLength, XtCStackLength, XtRInt, sizeof(int),
	offset(stack_len), XtRImmediate, (XtPointer) 0 },
    { XtNstack, XtCStack, XtRPointer, sizeof(XtPointer),
	offset(stack), XtRImmediate, (XtPointer) 0 },
    { XtNimageWidth, XtCImageWidth, XtRDimension, sizeof(Dimension),
	offset(width), XtRImmediate, (XtPointer) 0 },
    { XtNimageHeight, XtCImageHeight, XtRDimension, sizeof(Dimension),
	offset(height), XtRImmediate, (XtPointer) 0 }
#undef offset
};

static void Initialize __(( ImageWidget request, ImageWidget new ));
static XtGeometryResult
  QueryGeometry __(( ImageWidget, XtWidgetGeometry*, XtWidgetGeometry*));
static Boolean
  SetValues __(( ImageWidget w, ImageWidget request, ImageWidget new ));

static void ImRedisplay __(( ImageWidget w, XEvent *ev, Region region ));

ImageWidgetClassRec imageWidgetClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"Image",
    /* widget_size		*/	sizeof(ImageWidgetRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	XtInheritRealize,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	NULL,
    /* expose			*/	ImRedisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry		*/	QueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* image fields */
    /* empty			*/	0
  }
};

WidgetClass imageWidgetClass = (WidgetClass)&imageWidgetClassRec;

static void
Initialize _IMPL((request, new), ImageWidget request _AND ImageWidget new )
{
    XGCValues values;
    XtGCMask valueMask;
    ImageWidgetPart *newim = &new->image;

    newim->cvWidget = 0;
    newim->stack_editor = 0;
    newim->stack_total = newim->stack_len;
    newim->last_region = 0;

    /* Take over control of stack allocation. */
    if ( newim->stack ) {
	GenericDisplayParams *old = newim->stack;
	newim->stack = (GenericDisplayParams *)
	  XtCalloc(newim->stack_total, sizeof(GenericDisplayParams));
	bcopy(old, newim->stack,
	      newim->stack_total * sizeof(GenericDisplayParams));
    }

    /* Initialize image size from image on top of stack. */
    if ( newim->stack_len ) {
	GenericImage im = newim->stack[0]->generic.image;
	newim->width  = im->generic.width;
	newim->height = im->generic.height;

	if ( !new->core.width )
	  new->core.width  = (Dimension) rint(newim->mag_x * newim->width);
	if ( !new->core.height )
	  new->core.height = (Dimension) rint(newim->mag_y * newim->height);

	/* Install the initial top-of-stack image. */
	ImInstall(newim->stack[0]);
    }

    /* Size defaults. */
    if ( !new->core.width  ) new->core.width  = 200;
    if ( !new->core.height ) new->core.height = 200;

    valueMask = GCForeground | GCBackground;
    values.background = new->core.background_pixel;
    values.foreground = newim->foreground_pixel;
    newim->annotateGC = XtGetGC( (Widget)new, valueMask, &values );
}

static XtGeometryResult
QueryGeometry _IMPL((w, request, ret),
		    ImageWidget w _AND 
		    XtWidgetGeometry *request _AND XtWidgetGeometry *ret)
{
    XtGeometryResult result = XtGeometryYes;
    Dimension ideal_width, ideal_height;

    if ( w->image.stack_len > 0 ) {
	ideal_width  = (Dimension) rint(w->image.mag_x * w->image.width);
	ideal_height = (Dimension) rint(w->image.mag_y * w->image.height);

	if ( request->request_mode & CWWidth ) {
	    if ( request->width != ideal_width ) {
		ret->width = ideal_width;
		ret->request_mode |= CWWidth;
		result = XtGeometryAlmost;
	    }
	}
	if ( request->request_mode & CWHeight ) {
	    if ( request->height != ideal_height ) {
		ret->height = ideal_height;
		ret->request_mode |= CWHeight;
		result = XtGeometryAlmost;
	    }
	}
    }

    return result;
}

/*
 * Put the size of the smallest containing widget into the variables.
 */
static void
DisplaySize _IMPL((iw, width, height, dw),
		  ImageWidget iw _AND int *width _AND int *height _AND Widget *dw )
{
    Widget w = (Widget) iw;

    *width  = w->core.width;
    *height = w->core.height;
    if ( dw ) *dw = w;

    for ( ; w; w = w->core.parent ) {
	if ( w->core.width  < *width  ) {
	    *width  = w->core.width;
	    if ( dw ) *dw = w;
	}
	if ( w->core.height < *height ) {
	    *height = w->core.height;
	    if ( dw ) *dw = w;
	}
    }
}

static void
DisplayAnnotate _IMPL((w, region, range),
		      ImageWidget w _AND Region region _AND XRectangle *range )
{
    ImageWidgetPart *ipart = &w->image;
    float mag_x = ipart->mag_x;
    float mag_y = ipart->mag_y;
    GC gridGC   = ipart->annotateGC;
    Display *d  = XtDisplay(w);
    Window win  = XtWindow(w);
    XGCValues values;
    XtGCMask valueMask;
    
    Widget ww;
    int win_width, win_height;
    
    int i, j;
    
    if ( !ipart->grid && !ipart->descr ) return;

    /* 
     * The grid spacing may depend on the parent window (if it's 
     * smaller).
     */
    DisplaySize( w, &win_width, &win_height, &ww );
    
#define IM_X(i)  ((Position) rint(mag_x*(i)))
#define IM_Y(j)  ((Position) rint(mag_y*(j)))
    
    if ( ipart->grid ) {
	int grid_spacing, grid_highlight;

	for ( i = 0; i < XtNumber(grid_spacings); ++i ) {
	    grid_spacing   = grid_spacings[i];
	    grid_highlight = grid_highlights[i];
	    if (grid_spacing*mag_x >= MIN_PIXELS && 
		grid_spacing*mag_y >= MIN_PIXELS && 
		win_width /(mag_x*grid_spacing) <= MAX_GRID &&
		win_height/(mag_y*grid_spacing) <= MAX_GRID &&
		win_width /(mag_x*grid_highlight) >= MIN_HIGH &&
		win_height/(mag_y*grid_highlight) >= MIN_HIGH )
	      break;
	}
	
	/* Setup the GC for grid drawing. */
	valueMask = GCFont;
	values.font = ipart->grid_font->fid;
	XChangeGC( d, gridGC, valueMask, &values );
	
	for ( i = (range->x/grid_spacing)*grid_spacing; 
	     i <= range->x + range->width;
	     i += grid_spacing ) {
	    Boolean highlight = (i%grid_highlight==0);
	    if ( i == 0 && w->core.border_width != 0 ) continue;
	    XSetLineAttributes(d, gridGC, 0, 
			       (highlight? LineSolid: LineOnOffDash), 
			       CapButt, JoinMiter);
	    XDrawLine( d, win, gridGC,
		      IM_X(i), IM_Y(range->y),
		      IM_X(i), IM_Y(range->y+range->height) );
	    
	}
	for ( j = (range->y/grid_spacing)*grid_spacing; 
	     j <= range->y + range->height;
	     j += grid_spacing ) {
	    Boolean highlight = (j%grid_highlight==0);
	    if ( j == 0 && w->core.border_width != 0 ) continue;
	    XSetLineAttributes(d, gridGC, 0,
			       (highlight? LineSolid: LineOnOffDash), 
			       CapButt, JoinMiter);
	    XDrawLine( d, win, gridGC,
		      IM_X(range->x), IM_Y(j),
		      IM_X(range->x+range->width), IM_Y(j) );
	    /* When this guy's highlighted, then draw Coords descr. */
	    if ( highlight )
	      for ( i = (range->x/grid_spacing)*grid_spacing; 
		   i <= range->x + range->width;
		   i += grid_spacing ) {
		  if ( i%grid_highlight==0 ) {
		      char str[32];
		      sprintf( str, "%d,%d", i, j );
		      XDrawString(d, win, gridGC, 
				  IM_X(i) + 2, 
				  IM_Y(j) + 2 + w->image.grid_font->ascent, 
				  str, strlen(str));
		  }
	      }
	}
	
	XSetLineAttributes(d, gridGC, 0, LineSolid, CapButt, JoinMiter);
    }

    if ( ipart->descr ) {
	int x = 0, y = 0;
	char buf[BUFSIZ];

	/* Setup the GC for description. */
	valueMask = GCFont;
	values.font = ipart->descr_font->fid;
	XChangeGC( d, gridGC, valueMask, &values );
	XSetRegion(d, gridGC, region );

	if ( ww != (Widget) w ) {
	    x -= w->core.x;
	    y -= w->core.y;
	}
	x += 2;
	y += 2 + ipart->descr_font->ascent;

	for ( i = 0; i < MIN(ipart->displayN, ipart->stack_len); ++i ) {
	    ImDescription(ipart->stack[i]->generic.image, buf, BUFSIZ);
	    XDrawImageString(d, win, gridGC, x, y, buf, strlen(buf));

	    y += ipart->descr_font->ascent;
	    y += ipart->descr_font->descent;
	}
    }

#undef IM_X
#undef IM_Y
}

void
XParamsFromWidget _IMPL((xparams, w),
			ExternalParams *xparams _AND Widget w)
{
    xparams->widget = w;
    xparams->display = CMdisplay;
    xparams->screen = CMscreen;
    xparams->visual = CMvisual;
    xparams->win = XtWindow(w);
}

static void
DisplayParams _IMPL((w, region, range), 
		    ImageWidget w _AND Region region _AND XRectangle *range )
{
    ImageWidgetPart *ipart = &w->image;
    int i;

#define IM_X(i)  rint(mag_x*(i))
#define IM_Y(j)  rint(mag_y*(j))
#define IM_CX(i) rint(mag_x*((i)+0.5))
#define IM_CY(j) rint(mag_y*((j)+0.5))

    if ( ipart->stack_len ) {
	for ( i = MIN(ipart->displayN,ipart->stack_len) - 1; i >= 0; --i ) {
	    ExternalParams *xparams = ImExternalParams(ipart->stack[i]);
	    char buf[BUFSIZ];

	    if ( w->image.verbose ) 
	      printf( "Display %d: %#x %s\n", 
		     i, ipart->stack[i]->generic.image,
		     ImDescription(ipart->stack[i]->generic.image,buf,BUFSIZ) );
	    XParamsFromWidget(xparams, (Widget) w);
	    xparams->region = region;
	    xparams->mag_x = ipart->mag_x;
	    xparams->mag_y = ipart->mag_y;
	    xparams->range = *range;
	    
	    ImAllocColors( ipart->stack[i] );
	    ImDisplay( ipart->stack[i] );
	}
    }
#undef IM_X
#undef IM_Y
#undef IM_CX
#undef IM_CY
}

static void
RectRegion _IMPL((r, ret), Region r _AND Region *ret )
{
    XRectangle rect;

    if ( !*ret ) *ret = XCreateRegion();
    XClipBox( r, &rect );
    XUnionRectWithRegion( &rect, r, *ret );
}

static void
ImRedisplay _IMPL((w, ev, region),
		  ImageWidget w _AND XEvent *ev _AND Region region )
{
    {
	XRectangle clip;
	XClipBox(region, &clip);
	DEBUG((DBG_DISP, "Redisplay Ev(%d %d dpy=%#x win=%#x) @%d,%d %dx%d\n",
	       ev->type, ev->xany.serial, ev->xany.display, ev->xany.window,
	       clip.x, clip.y, clip.width, clip.height));
    }
    if ( w->core.visible && w->image.stack_len ) {
	Display *dpy = XtDisplay((Widget)w);
	Window win = XtWindow((Widget)w);
	Cursor watch = XCreateFontCursor(dpy, XC_watch);
	XRectangle clip_rect;
	Dimension width = w->image.width;
	Dimension height = w->image.height;

	RectRegion( region, &(w->image.last_region) );

	XDefineCursor(dpy,win,watch);
	XSync(XtDisplay(w),False);

	XClipBox( region, &clip_rect );

	clip_rect.x = (Position) floor(clip_rect.x/w->image.mag_x);
	clip_rect.y = (Position) floor(clip_rect.y/w->image.mag_y);
	clip_rect.width  = 
	  (Position) ceil( clip_rect.width /w->image.mag_x) + 1;
	if ( clip_rect.x + clip_rect.width > width ) {
	    if ( width <= clip_rect.x ) return;
	    clip_rect.width = width - clip_rect.x;
	}
	clip_rect.height = 
	  (Position) ceil( clip_rect.height/w->image.mag_y) + 1;
	if ( clip_rect.y + clip_rect.height > height ) {
	    if ( height <= clip_rect.y ) return;
	    clip_rect.height = height - clip_rect.y;
	}

	if ( !w->image.redraws_delayed )
	  DisplayParams( w, region, &clip_rect );
	DisplayAnnotate( w, region, &clip_rect );

	XUndefineCursor(dpy,win);
	XFreeCursor(dpy, watch);
    }
}

static Boolean
SetValues _IMPL((w, request, new),
		ImageWidget w _AND ImageWidget request _AND ImageWidget new )
{
    ImageWidgetPart 
      *oldim = &w->image, 
      *reqim = &request->image, 
      *newim = &new->image;
    Boolean resize = FALSE;

    if ( w->core.background_pixel != new->core.background_pixel ||
	 oldim->foreground_pixel != newim->foreground_pixel ) {
	XGCValues values;
	XtGCMask valueMask = GCBackground | GCForeground;
	values.background = w->core.background_pixel;
	values.foreground = newim->foreground_pixel;
	newim->annotateGC = XtGetGC((Widget)new, valueMask, &values);
    }
    if ( oldim->width != reqim->width ) {
	newim->width = reqim->width;
	resize = TRUE;
    }
    if ( oldim->height != reqim->height ) {
	newim->height = reqim->height;
	resize = TRUE;
    }
    if ( oldim->mag_x != reqim->mag_x ) {
	newim->mag_x = reqim->mag_x;
	resize = TRUE;
    }
    if ( oldim->mag_y != reqim->mag_y ) {
	newim->mag_y = reqim->mag_y;
	resize = TRUE;
    }

    if ( resize && newim->stack_len ) {
	new->core.width  = (Dimension) rint(newim->mag_x * newim->width);
	new->core.height = (Dimension) rint(newim->mag_y * newim->height);
    }

    if ( oldim->descr != reqim->descr || oldim->grid != reqim->grid )
      resize = TRUE;

    return resize;
}

void
ImDelayRedraws _IMPL((w), ImageWidget w )
{
    w->image.redraws_delayed = True;
}

void
ImRedraw _IMPL((w, force), ImageWidget w _AND Boolean force )
{
    if ( force ) w->image.redraws_delayed = False;
    DEBUG((DBG_DISP, "ImRedraw(%#x, %d) %s\n", w, force,
	   w->image.redraws_delayed? " Delayed": ""));
    if ( w->image.redraws_delayed )
      ++w->image.redraws_delayed;
    else if ( XtIsRealized((Widget) w) )
      XClearArea(XtDisplay((Widget)w), XtWindow((Widget)w), 0,0,0,0, True);
}


/* drag out a rectangle for selections etc */
static void
DrawBox(dpy, win, gc, x1, y1, x2, y2)
Display *dpy;
Window win;
GC gc;
int x1, y1, x2, y2;
{
    int x = MIN(x1,x2), y = MIN(y1,y2);
    int w = ABS(x1-x2), h = ABS(y1-y2);

    XDrawRectangle(dpy, win, gc, x, y, w, h);
}

typedef void 
  (*ImUpdateF) __((Widget w, int x1, int y1, int x2, int y2, Boolean ok));

static void
DragBox _IMPL((w, ret_x1, ret_y1, ret_x2, ret_y2, start_ev, update_f),
	      Widget w _AND 
	      int *ret_x1 _AND int *ret_y1 _AND 
	      int *ret_x2 _AND int *ret_y2 _AND 
	      XEvent *start_ev _AND ImUpdateF update_f)
{
#define DELAY_TIME 500
    Cursor ur, ul, ll, lr;          /* cursors for rubber banding    */
    int change_cursor = FALSE;
    int current_cursor;
    int x1, y1;                     /* location of mouse            */
    int x2, y2;                     /* other corner of box          */
    int rootx, rooty;		/* for XQueryPointer */
    unsigned int mask;	        /* for XQueryPointer          */
    Window root, subw;		/* for XQueryPointer */
    int chosen = -1;
    int stop;
    XEvent ev;                      /* someplace to put the event   */
    int events;                     /* what events we want.             */
    
    GC gc;
    XGCValues gcvalues;
    Window win;
    Display *dpy;
    int current_x, current_y;
    
    dpy = XtDisplay(w);
    win = XtWindow(w);

    events = PointerMotionMask | ButtonPressMask | ButtonReleaseMask;
    ul = XCreateFontCursor(dpy, XC_ul_angle);
    
    if (XGrabPointer(dpy, win, FALSE, events, 
		     GrabModeAsync, GrabModeAsync, None, ul, CurrentTime) != 
	GrabSuccess) {
	XtWarning("Can't Grab mouse in DragBox()");
	XFreeCursor(dpy, ul);
	return;
    }
    
    (void)XQueryPointer(dpy, win, &root, &subw, 
			&rootx, &rooty, &x1, &y1, &mask);
    
    /* make the GC that inverts the lines of a rectangle */
    XtVaGetValues(w, XtNbackground, &(gcvalues.foreground), NULL);
    gcvalues.function = GXinvert;
    gcvalues.line_width = 3;
    gc = XtGetGC(w, GCLineWidth|GCFunction|GCForeground, &gcvalues);
    
    *ret_x1 = *ret_y1 = *ret_x2 = *ret_y2 = 0;
    ur = XCreateFontCursor(dpy, XC_ur_angle);
    ll = XCreateFontCursor(dpy, XC_ll_angle);
    lr = XCreateFontCursor(dpy, XC_lr_angle);
    current_cursor = ul;
    
    current_x = x2 = x1; 
    current_y = y2 = y1;
    stop = FALSE;
    
    if ( start_ev && start_ev->type == ButtonPress ) {
	chosen = start_ev->xbutton.button;
	current_x = x1 = x2 = start_ev->xbutton.x;
	current_y = y1 = y2 = start_ev->xbutton.y;
	change_cursor = TRUE;
	/* XGrabServer(dpy); */
	if ( update_f ) (*update_f)(w,x1,y1,x2,y2,change_cursor);
    }
    
    while (stop == FALSE) {
	XFlush(dpy);
	XNextEvent(dpy, &ev);
	
	switch (ev.type) {
	  case ButtonPress:
	    if (chosen != -1)
	      break;
	    chosen = ev.xbutton.button;
	    current_x = x1 = x2 = ev.xbutton.x;
	    current_y = y1 = y2 = ev.xbutton.y;
	    change_cursor = TRUE;
	    /* XGrabServer(dpy); */
	    break;
	  case ButtonRelease:
	    if (ev.xbutton.button != chosen)
	      break;
	    x2 = ev.xbutton.x;
	    y2 = ev.xbutton.y;
	    if (change_cursor) {
		/* XUngrabServer(dpy); */
		DrawBox(dpy, win, gc, x1, y1, x2, y2);
	    }
	    stop = TRUE;
	    break;
	  case MotionNotify:
	    if (XPending(dpy)) {
		XEvent pev;
		XPeekEvent(dpy, &pev);
		while (pev.type == MotionNotify) {
		    XNextEvent(dpy, &ev);
		    if (XPending(dpy))
		      XPeekEvent(dpy, &pev);
		    else
		      break;
		};
	    }
	    x2 = ev.xmotion.x;
	    y2 = ev.xmotion.y;
	    break;
	  case KeyPress:
	    x2 = ev.xbutton.x;
	    y2 = ev.xbutton.y;
	    if (change_cursor) {
		/* XUngrabServer(dpy); */
		DrawBox(dpy, win, gc, x1, y1, x2, y2);
	    }
	    /* The abort conditions. */
	    x1 = x2 = current_x;
	    y1 = y2 = current_y;
	    stop = TRUE;
	    break;
	  default:
	    break;
	}
	if (current_x != x2 || current_y != y2) {
	    if (change_cursor)
	      DrawBox(dpy, win, gc, x1, y1, current_x, current_y);
	    if (change_cursor)
	      DrawBox(dpy, win, gc, x1, y1, x2, y2);
	}
	if ( update_f ) (*update_f)(w,x1,y1,x2,y2,change_cursor);

	current_x = x2;  current_y = y2;
	if (change_cursor) {
	    if ((x2 >= x1) && (y2 >= y1) && current_cursor != lr) {
		XChangeActivePointerGrab(dpy, events, lr, CurrentTime);
		current_cursor = lr;
	    } else if ((x2 >= x1) && (y2 < y1) && current_cursor != ur) {
		XChangeActivePointerGrab(dpy, events, ur, CurrentTime);
		current_cursor = ur;
	    } else if ((x2 < x1) && (y2 >= y1) && current_cursor != ll) {
		XChangeActivePointerGrab(dpy, events, ll, CurrentTime);
		current_cursor = ll;
	    } else if ((x2 < x1) && (y2 < y1) && current_cursor != ul) {
		XChangeActivePointerGrab(dpy, events, ul, CurrentTime);
		current_cursor = ul;
	    }
	}
    }
    /* XUngrabServer(dpy); */
    XUngrabPointer(dpy, CurrentTime);
    XFreeCursor(dpy, ur);
    XFreeCursor(dpy, ul);
    XFreeCursor(dpy, lr);
    XFreeCursor(dpy, ll);
    XtReleaseGC(w, gc);
    if (change_cursor) {
	*ret_x1 = MIN(x1, x2);
	*ret_y1 = MIN(y1, y2);
	*ret_x2 = MAX(x1, x2);
	*ret_y2 = MAX(y1, y2);
    } else {
	*ret_x1 = *ret_x2 = x2;
	*ret_y1 = *ret_y2 = y2;
    }
    XFlush(dpy);
}

#define XK_MISCELLANY
#include <X11/keysymdef.h>

static void
DragLine _IMPL((w, ret_x1, ret_y1, ret_x2, ret_y2, start_ev, update_f),
	       Widget w _AND 
	       int *ret_x1 _AND int *ret_y1 _AND 
	       int *ret_x2 _AND int *ret_y2 _AND 
	       XEvent *start_ev _AND ImUpdateF update_f)
{
    Cursor arrow1, arrow2;
    int x1, y1;                     /* location of mouse */
    int x2, y2;                     /* other end of line */
    int rootx, rooty;		/* for XQueryPointer */
    unsigned int mask;		/* for XQueryPointer */
    Window root, subw;		/* for XQueryPointer */
    int chosen = -1;
    int stop;
    XEvent ev;                      /* someplace to put the event */
    KeySym keysym;
    int events;                     /* what events we want. */
    
    GC gc;
    XGCValues gcvalues;
    Window win;
    Display *dpy;
    Boolean ok2draw;
    int current_x, current_y;
    
    *ret_x1 = *ret_y1 = *ret_x2 = *ret_y2 = 0;
    dpy = XtDisplay(w);
    win = XtWindow(w);
    events = PointerMotionMask | ButtonPressMask | ButtonReleaseMask;
    arrow1 = XCreateFontCursor(dpy, XC_draft_large);
    
    if (XGrabPointer(dpy, win, FALSE, events, 
		     GrabModeAsync, GrabModeAsync, None, 
		     arrow1, CurrentTime) != GrabSuccess) {
	XtWarning("Can't Grab the mouse in DragLine()");
	XFreeCursor(dpy, arrow1);
	return;
    }
    
    (void)XQueryPointer(dpy, win, &root, &subw, 
			&rootx, &rooty, &x1, &y1, &mask);
    
    arrow2 = XCreateFontCursor(dpy, XC_top_left_arrow);
    
    /* make the GC that inverts the line */
    XtVaGetValues(w, XtNbackground, &(gcvalues.foreground), NULL);
    gcvalues.function = GXinvert;
    gcvalues.line_width = 3;
    gc = XtGetGC(w, GCLineWidth|GCFunction|GCForeground, &gcvalues);
    
    current_x = x2 = x1; 
    current_y = y2 = y1;
    stop = FALSE;
    ok2draw = FALSE;

    if ( start_ev && start_ev->type == ButtonPress ) {
	chosen = start_ev->xbutton.button;
	current_x = x1 = x2 = start_ev->xbutton.x;
	current_y = y1 = y2 = start_ev->xbutton.y;
	ok2draw = TRUE;
	XDrawLine(dpy, win, gc, x1, y1, x2, y2);
	XChangeActivePointerGrab(dpy, events, arrow2, CurrentTime);
	/* XGrabServer(dpy); */
	if ( update_f ) (*update_f)(w,x1,y1,x2,y2,ok2draw);
    }

    while (stop == FALSE) {
	XFlush(dpy);
	XNextEvent(dpy, &ev);
	
	switch (ev.type) {
	  case ButtonPress:
	    if (chosen != -1) break;
	    chosen = ev.xbutton.button;
	    current_x = x1 = x2 = ev.xbutton.x;
	    current_y = y1 = y2 = ev.xbutton.y;
	    ok2draw = TRUE;
	    XDrawLine(dpy, win,  gc, x1, y1, x2, y2);
	    XChangeActivePointerGrab(dpy, events, arrow2, CurrentTime);
	    /* XGrabServer(dpy); */
	    break;
	  case ButtonRelease:
	    if (ev.xbutton.button != chosen) break;
	    x2 = ev.xbutton.x;
	    y2 = ev.xbutton.y;
	    if ( ev.xmotion.state & ShiftMask ) {
		if ( ABS(x1-x2) > ABS(y1-y2) ) y2 = y1;
		else x2 = x1;
	    }
	    if (ok2draw) {
		/* XUngrabServer(dpy); */
		XDrawLine(dpy, win, gc, x1, y1, x2, y2);
	    }
	    stop = TRUE;
	    break;
	  case MotionNotify:
	    if (XPending(dpy)) {
		XEvent pev;
		XPeekEvent(dpy, &pev);
		while (pev.type == MotionNotify) {
		    XNextEvent(dpy, &ev);
		    if (XPending(dpy))
		      XPeekEvent(dpy, &pev);
		    else
		      break;
		};
	    }
	    x2 = ev.xmotion.x;
	    y2 = ev.xmotion.y;
	    if ( ev.xmotion.state & ShiftMask ) {
		if ( ABS(x1-x2) > ABS(y1-y2) ) y2 = y1;
		else x2 = x1;
	    }
	    break;
	  case KeyPress:
	    keysym = XLookupKeysym((XKeyEvent*)&ev,0);
	    if ( IsModifierKey(keysym) ) break;
	    x2 = ev.xbutton.x;
	    y2 = ev.xbutton.y;
	    if (ok2draw) {
		/* XUngrabServer(dpy); */
		XDrawLine(dpy, win, gc, x1, y1, x2, y2);
	    }
	    x1 = x2 = current_x;
	    y1 = y2 = current_y;
	    stop = TRUE;
	    break;
	  default:
	    break;
	}
	if ( update_f ) (*update_f)(w,x1,y1,x2,y2,ok2draw);
	if (current_x != x2 || current_y != y2) {
	    if (ok2draw)
	      XDrawLine(dpy, win, gc, x1, y1, current_x, current_y);
	    if (ok2draw)
	      XDrawLine(dpy, win, gc, x1, y1, x2, y2);
	    current_x = x2;
	    current_y = y2;
	}
    }

    /* XUngrabServer(dpy); */
    XUngrabPointer(dpy, CurrentTime);
    XFreeCursor(dpy, arrow1);
    XFreeCursor(dpy, arrow2);
    XtReleaseGC(w, gc);

    *ret_x1 = x1;
    *ret_y1 = y1;
    *ret_x2 = x2;
    *ret_y2 = y2;

    XFlush(dpy);
}

static Boolean
DragPoint _IMPL((w, ret_x1, ret_y1, start_ev, update_f),
		Widget w _AND int *ret_x1 _AND int *ret_y1 _AND 
		XEvent *start_ev _AND ImUpdateF update_f)
{
    Cursor cross;
    int x1, y1;                     /* location of mouse */
    int rootx, rooty;		/* for XQueryPointer */
    unsigned int mask;		/* for XQueryPointer */
    Window root, subw;		/* for XQueryPointer */
    int chosen = -1;
    int stop;
    XEvent ev;                      /* someplace to put the event */
    int events;                     /* what events we want. */
    
    GC gc;
    XGCValues gcvalues;
    Window win;
    Display *dpy;
    int current_x, current_y;
    Boolean success = FALSE, ok2draw = FALSE;
    
    *ret_x1 = *ret_y1 = 0;
    dpy = XtDisplay(w);
    win = XtWindow(w);
    events = PointerMotionMask | ButtonPressMask | ButtonReleaseMask;
    cross = XCreateFontCursor(dpy, XC_crosshair);
    
    if (XGrabPointer(dpy, win, FALSE, events, 
		     GrabModeAsync, GrabModeAsync, None, 
		     cross, CurrentTime) != GrabSuccess) {
	XtWarning("Can't Grab the mouse in DragPoint()");
	XFreeCursor(dpy, cross);
	return FALSE;
    }
    
    (void)XQueryPointer(dpy, win, &root, &subw, 
			&rootx, &rooty, &x1, &y1, &mask);
    
    /* make the GC that inverts the line */
    XtVaGetValues(w, XtNbackground, &(gcvalues.foreground), NULL);
    gcvalues.function = GXinvert;
    gcvalues.line_width = 3;
    gc = XtGetGC(w, GCFunction|GCForeground|GCLineWidth, &gcvalues);
    
    current_x = x1; 
    current_y = y1;
    stop = FALSE;

    if ( start_ev && start_ev->type == ButtonPress ) {
	chosen = start_ev->xbutton.button;
	current_x = x1  = start_ev->xbutton.x;
	current_y = y1  = start_ev->xbutton.y;
	ok2draw = TRUE;
	XDrawLine(dpy, win, gc, x1-10, y1, x1+10, y1);
	XDrawLine(dpy, win, gc, x1, y1-10, x1, y1+10);
	/* XGrabServer(dpy); */
	if ( update_f ) (*update_f)(w,x1,y1,current_x,current_y,ok2draw);
    }

    while (stop == FALSE) {
	XFlush(dpy);
	XNextEvent(dpy, &ev);
	
	switch (ev.type) {
	  case ButtonPress:
	    if (chosen != -1) break;
	    chosen = ev.xbutton.button;
	    current_x = x1 = ev.xbutton.x;
	    current_y = y1 = ev.xbutton.y;
	    ok2draw = TRUE;
	    XDrawLine(dpy, win, gc, x1-10, y1, x1+10, y1);
	    XDrawLine(dpy, win, gc, x1, y1-10, x1, y1+10);
	    /* XGrabServer(dpy); */
	    break;
	  case ButtonRelease:
	    if (ev.xbutton.button != chosen) break;
	    current_x = ev.xbutton.x;
	    current_y = ev.xbutton.y;
	    if (ok2draw) {
		/* XUngrabServer(dpy); */
		XDrawLine(dpy, win, gc, x1-10, y1, x1+10, y1);
		XDrawLine(dpy, win, gc, x1, y1-10, x1, y1+10);
	    }
	    success = ABS(current_x-x1)<=10 &&  ABS(current_y-y1)<=10;
	    stop = TRUE;
	    break;
	  case MotionNotify:
	    if (XPending(dpy)) {
		XEvent pev;
		XPeekEvent(dpy, &pev);
		while (pev.type == MotionNotify) {
		    XNextEvent(dpy, &ev);
		    if (XPending(dpy))
		      XPeekEvent(dpy, &pev);
		    else
		      break;
		};
	    }
	    current_x = ev.xmotion.x;
	    current_x = ev.xmotion.y;
	    break;
	  case KeyPress:
	    if (ok2draw) {
		/* XUngrabServer(dpy); */
		XDrawLine(dpy, win, gc, x1-10, y1, x1+10, y1);
		XDrawLine(dpy, win, gc, x1, y1-10, x1, y1+10);
	    }
	    success = FALSE;
	    stop = TRUE;
	    break;
	  default:
	    break;
	}
    }

    /* XUngrabServer(dpy); */
    XUngrabPointer(dpy, CurrentTime);
    XFreeCursor(dpy, cross);
    XtReleaseGC(w, gc);

    *ret_x1 = x1;
    *ret_y1 = y1;

    XFlush(dpy);
    return success;
}

Boolean
ImSelectBox _IMPL((w, x1, y1, x2, y2, sev),
		  ImageWidget w _AND 
		  int *x1 _AND int *y1 _AND int *x2 _AND int *y2 _AND
		  XEvent *sev )
{
    float mag_x = w->image.mag_x;
    float mag_y = w->image.mag_y;
    XtAppContext app_context =
      XtWidgetToApplicationContext((Widget)w);

    if ( !w->image.stack_len ) return False;

    /* Flush events before selecting. */
    while ( XtAppPending(app_context) ) {
	XEvent event;
	XtAppNextEvent(app_context,&event);
	XtDispatchEvent(&event);
    }

    /* Select the box in question. */
    DragBox( (Widget)w, x1,y1, x2,y2, sev, NULL);
    
    *x1 /= mag_x; *y1 /= mag_y;
    *x2 /= mag_x; *y2 /= mag_y;

    if ( *x1 < 0 ) *x1 = 0;
    if ( *y1 < 0 ) *y1 = 0;
    if ( *x1 >= w->image.width ) *x1 = w->image.width-1;
    if ( *y1 >= w->image.height ) *y1 = w->image.height-1;

    return *x1!=*x2 && *y1!=*y2;
}

static void
ClipLine _IMPL((x1, y1, x2, y2, xmin, ymin, xmax, ymax),
	       int *x1 _AND int *y1 _AND int *x2 _AND int *y2 _AND
	       int xmin _AND int ymin _AND int xmax _AND int ymax )
{
    float slope = (*x1==*x2)? 0: (*y2-*y1)/(float)(*x2-*x1);

    /* X1 out of bounds. */
    if ( *x1 < xmin ) {
	if ( *x1==*x2 ) { *y2 = *y1; return; };
	*x1 = xmin;
	*y1 = *y2 - (*x2-*x1)*slope;
    } else if ( *x1 > xmax ) {
	if ( *x1==*x2 ) { *y2 = *y1; return; };
	*x1 = xmax;
	*y1 = *y2 - (*x2-*x1)*slope;
    }

    /* Y1 out of bounds. */
    if ( *y1 < ymin ) {
	if ( *y1==*y2 ) { *x2 = *x1; return; };
	*y1 = ymin;
	*x1 = *x2 - (*y2-*y1)/slope;
    } else if ( *y1 > ymax ) {
	if ( *y1==*y2 ) { *x2 = *x1; return; };
	*y1 = ymax;
	*x1 = *x2 - (*y2-*y1)/slope;
    }

    /* X2 out of bounds. */
    if ( *x2 < xmin ) {
	if ( *x1==*x2 ) { *y2 = *y1; return; };
	*x2 = xmin;
	*y2 = *y1 + (*x2-*x1)*slope;
    } else if ( *x2 > xmax ) {
	if ( *x1==*x2 ) { *y2 = *y1; return; };
	*x2 = xmax;
	*y2 = *y1 + (*x2-*x1)*slope;
    }

    /* Y2 out of bounds. */
    if ( *y2 < ymin ) {
	if ( *y1==*y2 ) { *x2 = *x1; return; };
	*y2 = ymin;
	*x2 = *x1 + (*y2-*y1)/slope;
    } else if ( *y2 > ymax ) {
	if ( *y1==*y2 ) { *x2 = *x1; return; };
	*y2 = ymax;
	*x2 = *x1 + (*y2-*y1)/slope;
    }
}

Boolean
ImSelectLine _IMPL((w, x1, y1, x2, y2, sev), 
		   ImageWidget w _AND 
		   int *x1 _AND int *y1 _AND int *x2 _AND int *y2 _AND 
		   XEvent *sev )
{
    float mag_x = w->image.mag_x;
    float mag_y = w->image.mag_y;
    XtAppContext app_context =
      XtWidgetToApplicationContext((Widget)w);

    if ( !w->image.stack_len ) return False;

    /* Flush events before selecting. */
    while ( XtAppPending(app_context) ) {
	XEvent event;
	XtAppNextEvent(app_context,&event);
	XtDispatchEvent(&event);
    }

    /* Select the line in question. */
    DragLine( (Widget)w, x1,y1, x2,y2, sev, NULL);
    
    *x1 /= mag_x; *y1 /= mag_y;
    *x2 /= mag_x; *y2 /= mag_y;

    ClipLine( x1,y1, x2,y2, 0,0, w->image.width-1, w->image.height-1 );

    return *x1!=*x2 || *y1!=*y2;
}

Boolean
ImSelectPoint _IMPL((w, x1, y1, sev),
		    ImageWidget w _AND int *x1 _AND int *y1 _AND XEvent *sev )
{
    float mag_x = w->image.mag_x;
    float mag_y = w->image.mag_y;
    XtAppContext app_context =
      XtWidgetToApplicationContext((Widget)w);

    if ( !w->image.stack_len ) return False;

    /* Flush events before selecting. */
    while ( XtAppPending(app_context) ) {
	XEvent event;
	XtAppNextEvent(app_context,&event);
	XtDispatchEvent(&event);
    }

    /* Select the point in question. */
    if ( DragPoint( (Widget)w, x1,y1, sev, NULL) ) {
        *x1 /= mag_x; *y1 /= mag_y;

	if ( *x1 < 0 ) *x1 = 0;
	if ( *y1 < 0 ) *y1 = 0;
	if ( *x1 >= w->image.width )  *x1 = w->image.width;
	if ( *y1 >= w->image.height ) *y1 = w->image.height;
	return True;
    }
    return False;
}
