/*
 * $XConsortium: popup.c,v 2.26 89/12/10 17:38:59 converse Exp $
 *
 *
 *			  COPYRIGHT 1989
 *		   DIGITAL EQUIPMENT CORPORATION
 *		       MAYNARD, MASSACHUSETTS
 *			ALL RIGHTS RESERVED.
 *
 * THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE AND
 * SHOULD NOT BE CONSTRUED AS A COMMITMENT BY DIGITAL EQUIPMENT CORPORATION.
 * DIGITAL MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR
 * ANY PURPOSE.  IT IS SUPPLIED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
 *
 * IF THE SOFTWARE IS MODIFIED IN A MANNER CREATING DERIVATIVE COPYRIGHT
 * RIGHTS, APPROPRIATE LEGENDS MAY BE PLACED ON THE DERIVATIVE WORK IN
 * ADDITION TO THAT SET FORTH ABOVE.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Digital Equipment Corporation not be
 * used in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.
 */

#include <varargs.h>
#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/AsciiText.h>

struct _LastInput {
    Window win;
    int x;
    int y;
} lastInput;

typedef struct _PopupStatus {
	Widget popup;
	struct _LastInput lastInput;
} PopupStatusRec, *PopupStatus;

extern Widget toplevel;

static void DeterminePopupPosition(x_ptr, y_ptr)
    Position	*x_ptr, *y_ptr;
{
    Arg		args[3];
    Widget	source;
    Dimension	width, height;

    if (lastInput.win != -1) 
	source = XtWindowToWidget( XtDisplay(toplevel), lastInput.win);
    else
	source = toplevel;	/* %%% need to keep track of last screen */

    if (source != (Widget)NULL) {
	XtSetArg( args[0], XtNwidth, &width );
	XtSetArg( args[1], XtNheight, &height );
	XtGetValues( source, args, TWO );
	XtTranslateCoords( source, (Position) (width / 2),
			  (Position) (height / 2), x_ptr, y_ptr);
    } else {
	*x_ptr = lastInput.x;
	*y_ptr = lastInput.y;
    }
}

static Boolean PositionThePopup(popup, x, y)
    Widget	popup;
    Position	x, y;
{
    /* Hack.  Fix up the position of the popup.  The xmh app defaults file
     * contains an Xmh*Geometry specification; the effects of that on 
     * popups, and the lack of any user-supplied geometry specification for
     * popups, are mitigated here, by giving the popup shell a position.
     * (Xmh*Geometry is needed in case there is no user-supplied default.)
     * Returns True if an explicit geometry was inferred; false if the
     * widget was repositioned to (x,y).
     */

    Arg		args[4];
    String 	top_geom, pop_geom;

    XtSetArg( args[0], XtNgeometry, &top_geom );
    XtGetValues( toplevel, args, ONE );
    XtSetArg( args[0], XtNgeometry, &pop_geom );
    XtGetValues( popup, args, ONE );

    if (pop_geom == NULL || pop_geom == top_geom) {
	/* if same db entry, then ... */
	XtSetArg( args[0], XtNgeometry, (String) NULL);
	XtSetArg( args[1], XtNx, x);
	XtSetArg( args[2], XtNy, y);
	XtSetArg( args[3], XtNwinGravity, SouthWestGravity);
	XtSetValues( popup, args, FOUR);
	return False;
    }
    return True;
}


static void CenterPopupPosition(center, popup, px, py)
    Widget	center;
    Widget	popup;
    Position	*px, *py;
{
    Position	x, y;
    Position	nx, ny;
    Dimension	pw, ph;
    Arg		args[3];

    XtVaGetValues(popup,XtNwidth,&pw,XtNheight,&ph,XtNx,&x,XtNy,&y,NULL);

    if (center == NULL) {
	/* Center over pointer position. */
	Window r, c;
	int rx, ry, cx, cy;
	unsigned int m;

	XQueryPointer(XtDisplay(popup), XtWindow(popup), 
		      &r,&c, &rx,&ry,&cx,&cy, &m);
	*px = x = rx-pw/2; if ( x < 0 ) x = 0;
	*py = y = ry-ph/2; if ( y < 0 ) y = 0;
	XtVaSetValues(popup, XtNx, x, XtNy, y, NULL);
	return;
    }

    if (x == *px && y == *py) {

	/* Program sets geometry.  Correct our earlier calculations. */
	Dimension ww, wh;
	XtVaGetValues(center,XtNwidth,&ww,XtNheight,&wh,NULL);
	nx = (ww-pw) / 2;
	ny = (wh-ph) / 2;
	if (nx < 0) nx = 0;
	if (ny < 0) ny = 0;
	XtTranslateCoords(center, nx, ny, &x, &y);
	XtSetArg(args[0], XtNx, x);
	XtSetArg(args[1], XtNy, y);
	XtSetArg(args[2], XtNwinGravity, CenterGravity);
	XtSetValues(popup, args, THREE);
    }
}
	 

/* Ensure that the popup is wholly showing on the screen.
   Optionally center the widget horizontally and/or vertically
   on current position.
 */

static void EnsureVisibility(popup, popup_child, x, y, centerX, centerY)
    Widget	popup, popup_child;
    Position	x, y;		/* assert: current position = (x,y) */
    Boolean	centerX, centerY;
{
    Position	root_x, root_y;
    Dimension	width, height, border;
    Arg		args[3];


    XtSetArg( args[0], XtNwidth, &width );
    XtSetArg( args[1], XtNheight, &height );
    XtSetArg( args[2], XtNborderWidth, &border );
    XtGetValues( popup, args, THREE );

    XtTranslateCoords(popup_child, (Position)0, (Position)0, &root_x, &root_y);
    if (centerX) root_x -= width/2 + border;
    if (centerY) root_y -= height/2 + border;
    if (root_x < 0) root_x = 0;
    if (root_y < 0) root_y = 0;
    border <<= 1;

    if ((root_x + width + border) > WidthOfScreen(XtScreen(toplevel))) {
	root_x = WidthOfScreen(XtScreen(toplevel)) - width - border;
    }
    if ((root_y + height + border) > HeightOfScreen(XtScreen(toplevel))) {
	root_y = HeightOfScreen(XtScreen(toplevel)) - height - border;
    }

    if (root_x != x || root_y != y) {
	XtSetArg( args[0], XtNx, root_x );
	XtSetArg( args[1], XtNy, root_y );
	XtSetValues( popup, args, TWO );
    }
}


/*ARGSUSED*/
void DestroyPopup(widget, client_data, call_data)
    Widget		widget;		/* unused */
    XtPointer		client_data;
    XtPointer		call_data;	/* unused */
{
    Widget		popup = (Widget) client_data;
    XtPopdown(popup);
    XtDestroyWidget(popup);
    XSync(XtDisplay(widget),False);
}


#define OKAY_NAME "okay"

/*ARGSUSED*/
void XmhPromptOkayAction(w, event, params, num_params)
    Widget	w;		/* the "value" widget in the Dialog box */
    XEvent	*event;		/* unused */
    String	*params;	/* unused */
    Cardinal	*num_params;	/* unused */
{
    XtCallCallbacks(XtNameToWidget(XtParent(w), OKAY_NAME), XtNcallback,
		    XtParent(w));
}


void PopupPrompt(va_alist)
va_dcl
{
    va_list		ap;
    String		format;
    XtCallbackProc	okayCallback;		/* CreateFolder() */
    char		question[128];		/* the prompting string */
    Arg			args[4];
    Widget		popup;
    Widget		dialog;
    Widget		value;
    Position		x, y;
    Boolean		positioned;

    static String text_translations =
	"<Key>Return: XmhPromptOkayAction()\n\
         Ctrl<Key>R:  no-op(RingBell)\n\
         Ctrl<Key>S:  no-op(RingBell)\n";

    va_start(ap);
    okayCallback = va_arg(ap,XtCallbackProc);
    format = va_arg(ap,String);
    vsprintf(question,format,ap);
    va_end(ap);

    DeterminePopupPosition(&x, &y);
    XtSetArg(args[0], XtNallowShellResize, True);
    XtSetArg(args[1], XtNinput, True);
    popup = XtCreatePopupShell("prompt", transientShellWidgetClass, toplevel,
			       args, TWO);
    positioned = PositionThePopup(popup, x, y);

    XtSetArg(args[0], XtNlabel, question);
    XtSetArg(args[1], XtNvalue, "");
    dialog = XtCreateManagedWidget("dialog", dialogWidgetClass, popup, args,
				   TWO);
    XtSetArg(args[0], XtNresizable, True);
    XtSetValues( XtNameToWidget(dialog, "label"), args, ONE);
    value = XtNameToWidget(dialog, "value");
    XtSetValues( value, args, ONE);
    XtOverrideTranslations(value, XtParseTranslationTable(text_translations));

    XawDialogAddButton(dialog, OKAY_NAME, okayCallback, (XtPointer) dialog);
    XawDialogAddButton(dialog, "cancel", DestroyPopup, (XtPointer) popup);
    XtInstallAllAccelerators(popup, popup);
    XtRealizeWidget(popup);
    EnsureVisibility(popup, dialog, x, y, positioned ? False : True, False);
    XtPopup(popup, XtGrabNone);
}

#undef OKAY_NAME

/* ARGSUSED */
static void FreePopupStatus( w, closure, call_data )
    Widget w;			/* unused */
    XtPointer closure;
    XtPointer call_data;	/* unused */
{
    PopupStatus popup = (PopupStatus)closure;
    XtPopdown(popup->popup);
    XtDestroyWidget(popup->popup);
    XSync(XtDisplay(popup->popup),False);
    XtFree((char *) closure);
}


void PopupNotice(va_alist)
    va_dcl
{
    va_list ap;
    XtCallbackProc	callback;
    XtPointer		closure;
    char*		format;
    PopupStatus popup_status;
    char message[128];
    Arg args[5];
    Widget dialog;
    Widget value;
    Position x, y;
    char command[65], label[128];

    /* Collect arguments. */
    va_start(ap);
    callback = va_arg(ap,XtCallbackProc);
    closure = va_arg(ap,XtPointer);
    format = va_arg(ap,String);
    va_end(ap);

    /* Build actual message. */
    vsprintf(message,format,ap);

    popup_status = (PopupStatus)closure;
    if (popup_status == (PopupStatus)NULL) {
	popup_status = XtNew(PopupStatusRec);
	popup_status->lastInput = lastInput;
    }
    if (sscanf( message, "%64s", command ) != 1)
	(void) strcpy( command, "system" );
    else {
	int l = strlen(command);
	if (l && command[--l] == ':')
	    command[l] = '\0';
    }
    (void) sprintf( label, "%.64s command returned:", command );

    DeterminePopupPosition(&x, &y);
    XtSetArg( args[0], XtNallowShellResize, True );
    XtSetArg( args[1], XtNinput, True );
    popup_status->popup = XtCreatePopupShell( "notice",
			     transientShellWidgetClass, toplevel, args, TWO );
    PositionThePopup(popup_status->popup, x, y);

    XtSetArg( args[0], XtNlabel, label );
    XtSetArg( args[1], XtNvalue, message );
    dialog = XtCreateManagedWidget( "dialog", dialogWidgetClass,
				   popup_status->popup, args, TWO);

    /* The text area of the dialog box will not be editable. */
    value = XtNameToWidget(dialog, "value");
    XtSetArg( args[0], XtNeditType, XawtextRead);
    XtSetArg( args[1], XtNdisplayCaret, False);
    XtSetValues( value, args, TWO);
    /* XtOverrideTranslations(value, NoTextSearchAndReplace); */

    XawDialogAddButton( dialog, "confirm",
		       ((callback != (XtCallbackProc) NULL)
		          ? callback : (XtCallbackProc) FreePopupStatus), 
		       (XtPointer) popup_status
		      );

    XtRealizeWidget( popup_status->popup );
    XtInstallAllAccelerators(popup_status->popup, popup_status->popup);
    EnsureVisibility(popup_status->popup, dialog, x, y, False, False);
    XtPopup(popup_status->popup, XtGrabNone);
}


void PopupConfirm(va_alist)
va_dcl
{
    va_list	ap;
    Widget		center_widget;	/* where to center; may be NULL */
    XtCallbackList	affirm_callbacks;
    XtCallbackList	negate_callbacks;
    String		format;
    char	question[128];
    Arg		args[2];
    Widget	popup;
    Widget	dialog;
    Widget	button;
    Position	x, y;
    static XtCallbackRec callbacks[] = {
	{DestroyPopup,		(XtPointer) NULL},
	{(XtCallbackProc) NULL,	(XtPointer) NULL}
    };
    static Arg	shell_args[] = {
	{ XtNallowShellResize,	(XtArgVal) True},
	{ XtNinput,		(XtArgVal) True},
    };

    va_start(ap);
    center_widget = (Widget) va_arg(ap,char *);
    affirm_callbacks = va_arg(ap,XtCallbackList);
    negate_callbacks = va_arg(ap,XtCallbackList);
    format = va_arg(ap,String);
    vsprintf(question,format,ap);
    va_end(ap);

    DeterminePopupPosition(&x, &y);
    popup = XtCreatePopupShell("confirm", transientShellWidgetClass,
			       toplevel, shell_args, XtNumber(shell_args));
    PositionThePopup(popup, x, y); 

    XtSetArg(args[0], XtNlabel, question);
    dialog = XtCreateManagedWidget("dialog", dialogWidgetClass, popup, args,
				   ONE);
    
    callbacks[0].closure = (XtPointer) popup;
    XtSetArg(args[0], XtNcallback, callbacks);
    button = XtCreateManagedWidget("yes", commandWidgetClass, dialog, 
				   args, ONE);
    if (affirm_callbacks)
	XtAddCallbacks(button, XtNcallback, affirm_callbacks);


    button = XtCreateManagedWidget("no", commandWidgetClass, dialog, 
				   args, ZERO);
    XtAddCallback(button, XtNcallback, DestroyPopup, (XtPointer) popup);
    if (negate_callbacks)
	XtAddCallbacks(button, XtNcallback, negate_callbacks);

    XtRealizeWidget(popup);
    XtInstallAllAccelerators(popup, popup);
    CenterPopupPosition(center_widget, popup, &x, &y);
    EnsureVisibility(popup, dialog, x, y, False, False);
    XtPopup(popup, XtGrabNone);
}

static enum { ModalConfirm, ModalReject, ModalNotDone } modalStatus;

static void ConfirmModalPopup(widget, client_data, call_data)
    Widget		widget;		/* unused */
    XtPointer		client_data;
    XtPointer		call_data;	/* unused */
{
    Widget		popup = (Widget) client_data;
    modalStatus = ModalConfirm;
}

static void RejectModalPopup(widget, client_data, call_data)
    Widget		widget;		/* unused */
    XtPointer		client_data;
    XtPointer		call_data;	/* unused */
{
    Widget		popup = (Widget) client_data;
    modalStatus = ModalReject;
}

Boolean ModalPopupConfirm(va_alist)
va_dcl
{
    va_list	ap;
    String	format;
    char	question[128];
    Arg		args[2];
    Widget	popup;
    Widget	dialog;
    Widget	button;
    Position	x = -1, y = -1;
    static XtCallbackRec callbacks[] = {
	{DestroyPopup,		(XtPointer) NULL},
	{(XtCallbackProc) NULL,	(XtPointer) NULL}
    };
    static Arg	shell_args[] = {
	{ XtNallowShellResize,	(XtArgVal) True},
	{ XtNinput,		(XtArgVal) True},
    };

    va_start(ap);
    format = va_arg(ap,String);
    vsprintf(question,format,ap);
    va_end(ap);

    popup = XtCreatePopupShell("confirm", transientShellWidgetClass,
			       toplevel, shell_args, XtNumber(shell_args));

    XtSetArg(args[0], XtNlabel, question);
    dialog = XtCreateManagedWidget("dialog", dialogWidgetClass, popup, args,
				   ONE);
    
    callbacks[0].closure = (XtPointer) popup;
    XtSetArg(args[0], XtNcallback, callbacks);
    button = XtCreateManagedWidget("yes", commandWidgetClass, dialog, 
				   args, ONE);
    XtAddCallback(button, XtNcallback, ConfirmModalPopup, (XtPointer) popup);

    button = XtCreateManagedWidget("no", commandWidgetClass, dialog, 
				   args, ONE);
    XtAddCallback(button, XtNcallback, RejectModalPopup, (XtPointer) popup);

    XtRealizeWidget(popup);
    XtInstallAllAccelerators(popup, popup);
    CenterPopupPosition(NULL,popup,&x,&y);
    EnsureVisibility(popup, dialog, x, y, False, False);
    XtPopup(popup, XtGrabNone);

    modalStatus = ModalNotDone;
    while ( modalStatus == ModalNotDone ) {
	XEvent event;
	XtAppNextEvent(XtWidgetToApplicationContext(popup),&event); 
        XtDispatchEvent(&event);
    }
    return modalStatus == ModalConfirm;
}

void PopupError(va_alist)
va_dcl
{
    va_list	ap;
    String	format;
    char	message[128];
    Arg		args[3];
    Widget	error_popup, dialog;
    Position	x = -1, y = -1;
    static XtCallbackRec callbacks[] = {
	{DestroyPopup,		(XtPointer) NULL},
	{(XtCallbackProc) NULL,	(XtPointer) NULL}
    };

    va_start(ap);
    format = va_arg(ap,String);
    vsprintf(message,format,ap);
    va_end(ap);

    XtSetArg(args[0], XtNallowShellResize, True);
    XtSetArg(args[1], XtNinput, True);
    error_popup = XtCreatePopupShell("error", transientShellWidgetClass,
				     toplevel, args, TWO);

    XtSetArg(args[0], XtNlabel, message);
    dialog = XtCreateManagedWidget("dialog", dialogWidgetClass, error_popup,
				   args, ONE);
    callbacks[0].closure = (XtPointer) error_popup;
    XtSetArg(args[0], XtNcallback, callbacks);
    XawDialogAddButton(dialog, "OK", DestroyPopup, (XtPointer) error_popup);
    
    XtRealizeWidget(error_popup);
    XtInstallAllAccelerators(error_popup, error_popup);
    CenterPopupPosition(NULL,error_popup,&x,&y);
    EnsureVisibility(error_popup, dialog, x, y, False, False);

    XBell(XtDisplay(error_popup),50);
    XtPopup(error_popup, XtGrabNone);
}

void
PlacePopup( w, placed )
Widget w;
Boolean placed;
{
    Dimension width, height, border;
    Position x, y;

    XtVaGetValues(w, 
		  XtNwidth, &width, XtNheight, &height, 
		  XtNborderWidth, &border, 
		  NULL); 

    /* 
     * Either initially place the window under the mouse, or move the 
     * mouse over the window.
     */
    if ( !placed ) {
	Window r, c;
	int rx, ry, cx, cy;
	unsigned int m;
	Dimension w_scr, h_scr;

	/* Centre the widget over the pointer. */
	XQueryPointer(XtDisplay(w), XtWindow(w), &r,&c, &rx,&ry,&cx,&cy, &m);
	x = rx - width/2;  
	y = ry - height/2;

	/* Ensure that it is wholly on screen. */
	w_scr = WidthOfScreen(XtScreen(w));
	h_scr = HeightOfScreen(XtScreen(w));

	if ( x < 0 ) 
	  x = 0;
	else if ( (x + width + border) > w_scr )
	  x = w_scr - width - border;

	if ( y < 0 ) 
	  y = 0;
	else if ( (y + height + border) > h_scr )
	  y = h_scr - height - border;

	XtVaSetValues(w, XtNx, x, XtNy, y, NULL);
    } else {
	XWarpPointer( XtDisplay(w), NULL, XtWindow(w), 0,0,0,0, 
		      width/2, height/2 );
    }
}
