/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

   $Id: RGBImageP.h,v 0.9.1.2 1991/11/17 19:29:04 leei Exp $

******************************************************************************/

#ifndef _RGBImageP_h_
#define _RGBImageP_h_

#include <vim/RGBImage.h>
#include <vim/GenImageP.h>
#include <vim/RGBMap.h>
#include <vim/gamma.h>
#include <vim/proto.h>

typedef struct _RGBImageClassRec {
    GenericImageClassPart generic;
} RGBImageClassRec;

#ifndef _byte
#define _byte
typedef unsigned char byte;
#endif

typedef struct {
    unsigned short depth;
    unsigned short pixlen;
    byte *data;
    Pixmap pm;
    Dimension pm_width;
    Dimension pm_height;
    RGBMap cmap;
    unsigned long *pixel_counts;
} RGBImagePart;

typedef struct _RGBImageRec {
    GenericImagePart generic;
    RGBImagePart     rgb;
} RGBImageRec;

typedef struct {
  GammaParams	gamma;
  Boolean       redo_gamma;
  Pixel*	index;
  time_t	color_timestamp;
} RGBDisplayParamsPart;

typedef struct _RGBDisplayParamsRec {
    ExternalParams xparams;
    GenericDisplayParamsPart generic;
    RGBDisplayParamsPart     rgb;
} RGBDisplayParamsRec;

#define declare_RGBImage(pref)\
  unsigned short _CONCAT(pref,_width);\
  unsigned short _CONCAT(pref,_height);\
  RGBMap *_CONCAT(pref,_cmap);\
  unsigned short _CONCAT(pref,_depth);\
  unsigned short _CONCAT(pref,_pixlen);\
  byte *_CONCAT(pref,_data)

#define use_RGBImage(im,pref)\
  unsigned short _CONCAT(pref,_width) = (im)->generic.width;\
  unsigned short _CONCAT(pref,_height) = (im)->generic.height;\
  RGBMap *_CONCAT(pref,_cmap) = &(im)->rgb.cmap;\
  unsigned short _CONCAT(pref,_depth) = (im)->rgb.depth;\
  unsigned short _CONCAT(pref,_pixlen) = (im)->rgb.pixlen;\
  byte *_CONCAT(pref,_data) = (im)->rgb.data

#define access_RGBImage(im,pref)\
  _CONCAT(pref,_width) = (im)->generic.width;\
  _CONCAT(pref,_height) = (im)->generic.height;\
  _CONCAT(pref,_cmap) = &(im)->rgb.cmap;\
  _CONCAT(pref,_depth) = (im)->rgb.depth;\
  _CONCAT(pref,_pixlen) = (im)->rgb.pixlen;\
  _CONCAT(pref,_data) = (im)->rgb.data

#define end_RGBImage(im,pref)\
  (im)->generic.width = _CONCAT(pref,_width);\
  (im)->generic.height = _CONCAT(pref,_height);\
  (im)->rgb.depth = _CONCAT(pref,_depth);\
  (im)->rgb.pixlen = _CONCAT(pref,_pixlen);\
  (im)->rgb.data = _CONCAT(pref,_data)

#endif /* _RGBImageP_h_ */
