/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

   $Id: pFile.h,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $

******************************************************************************/

#ifndef _pFile_h_
#define _pFile_h_

#include <stdio.h>

#include <vim/proto.h>

typedef struct {
    FILE *fp;
    int pblen;
    unsigned char *pbuf;
    unsigned char *pb_l, *pb_e;
} *pFILE;

extern _SIGNED char _pfcharTable[];

pFILE pfopen __((const char *fname));
pFILE pfreopen __((FILE *fp));
pFILE pfstropen __((char *s, int len));
int pfclose __((pFILE));

#define pfeof(pf)	((pf->pb_e==pf->pb_l) && (!pf->fp || feof(pf->fp)))
#define pfgetc(pf)	((pf->pb_e<pf->pb_l)? (int) *(pf->pb_e++): \
			 (pf->fp? fgetc(pf->fp): EOF))

#define pfcharval(c)	(_pfcharTable[c])

int pfpeek __((void *ptr, int size, int n, pFILE));
int pfread __((void *ptr, int size, int n, pFILE));

char *pfgets __((char *s, int n, pFILE));
char *pfgetw __((char *s, int n, pFILE));

int pfgetint __((int *val, pFILE));
int pfgetuint __((unsigned int *val, pFILE));
int pfgetlint __((long *val, pFILE));
int pfgetulint __((unsigned long *val, pFILE));
int pfgethint __((short *val, pFILE));
int pfgetuhint __((unsigned short *val, pFILE));
int pfgetcint __((char *val, pFILE));
int pfgetucint __((unsigned char *val, pFILE));

void pfungetc __((char c, pFILE));
int pfungets __((char *s, pFILE));
int pfungetm __((void *ptr, int size, int n, pFILE));

int pfnextc __((pFILE pf));

int pfpushback __((char **buf, pFILE pf));

#include <rpc/rpc.h>

void xdr_pFILE_create __((XDR *xdrs, pFILE file, enum xdr_op op));

#endif /* _pFile_h_ */
