/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: xdr_pFile.c,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $";

/*
 * xdr_pFILE.c, XDR implementation on pushback file.
 *
 * This set of routines implements a XDR on a pFILE stream.
 * XDR_DECODE de-serializes from the stream.
 */

#include <rpc/rpc.h>
#include <vim/pFile.h>

static bool_t	xdrpFILE_getlong();
static bool_t	xdrpFILE_getbytes();
static long *	xdrpFILE_inline();
static void	xdrpFILE_destroy();

/*
 * Ops vector for pFILE type XDR
 */
static struct xdr_ops	xdrpFILE_ops = {
	xdrpFILE_getlong,	/* deserialize a long int */
	NULL,			/* serialize a long int */
	xdrpFILE_getbytes,	/* deserialize counted bytes */
	NULL,			/* serialize counted bytes */
	NULL,			/* get offset in the stream */
	NULL,			/* set offset in the stream */
	xdrpFILE_inline,	/* prime stream for inline macros */
	xdrpFILE_destroy	/* destroy stream */
};

/*
 * Initialize a pFILE xdr stream.
 * Sets the xdr stream handle xdrs for use on the stream file.
 * Operation flag is set to op.
 */
void
xdr_pFILE_create(xdrs, file, op)
	register XDR *xdrs;
	pFILE file;
	enum xdr_op op;
{
    if ( op == XDR_ENCODE ) {
	fprintf(stderr, 
		"xdr_pFILE_create(): Cannot ENCODE.  pFILE is read only.\n");
	abort();
    }
	
    xdrs->x_op = op;
    xdrs->x_ops = &xdrpFILE_ops;
    xdrs->x_private = (caddr_t)file;
    xdrs->x_handy = 0;
    xdrs->x_base = 0;
}

/*
 * Destroy a pFILE xdr stream.
 * Cleans up the xdr stream handle xdrs previously set up by xdrpFILE_create.
 */
static void
xdrpFILE_destroy(xdrs)
	register XDR *xdrs;
{
	/* xx should we close the file ?? */
};

static bool_t
xdrpFILE_getlong(xdrs, lp)
	XDR *xdrs;
	register long *lp;
{

	if (pfread((caddr_t)lp, sizeof(long), 1, 
		   (pFILE)xdrs->x_private) != 1)
	  return (FALSE);
#ifndef mc68000
	*lp = ntohl(*lp);
#endif
	return (TRUE);
}

static bool_t
xdrpFILE_getbytes(xdrs, addr, len)
	XDR *xdrs;
	caddr_t addr;
	u_int len;
{

	if ((len != 0) && 
	    (pfread(addr, (int)len, 1, (pFILE)xdrs->x_private) != 1))
	  return (FALSE);
	return (TRUE);
}

static long *
xdrpFILE_inline(xdrs, len)
	XDR *xdrs;
	u_int len;
{

	/*
	 * Must do some work to implement this: must insure
	 * enough data in the underlying pFILE buffer,
	 * that the buffer is aligned so that we can indirect through a
	 * long *, and stuff this pointer in xdrs->x_buf.  Doing
	 * a fread or fwrite to a scratch buffer would defeat
	 * most of the gains to be had here and require storage
	 * management on this buffer, so we don't do this.
	 */
	return (NULL);
}
